from rlcard.utils.utils import rank2int

class LeducholdemJudger:
    ''' The Judger class for Leduc Hold'em
    '''
    def __init__(self, np_random):
        ''' Initialize a judger class
        '''
        self.np_random = np_random

    @staticmethod
    def judge_game(players, public_card):
        ''' Judge the winner of the game.

        Args:
            players (list): The list of players who play the game
            public_card (object): The public card that seen by all the players

        Returns:
            (list): Each entry of the list corresponds to one entry of the
        '''
        # Judge who are the winners
        winners = [0] * len(players)
        fold_count = 0
        ranks = []
        # If every player folds except one, the alive player is the winner
        for idx, player in enumerate(players):
            ranks.append(rank2int(player.hand.rank))
            if player.status == 'folded':
               fold_count += 1
            elif player.status == 'alive':
                alive_idx = idx
        if fold_count == (len(players) - 1):
            winners[alive_idx] = 1
        
        # If any of the players matches the public card wins
        if sum(winners) < 1:
            for idx, player in enumerate(players):
                if player.hand.rank == public_card.rank:
                    winners[idx] = 1
                    break
        
        # If non of the above conditions, the winner player is the one with the highest card rank
        if sum(winners) < 1:
            max_rank = max(ranks)
            max_index = [i for i, j in enumerate(ranks) if j == max_rank]
            for idx in max_index:
                winners[idx] = 1

        # Compute the total chips
        total = 0
        for p in players:
            total += p.in_chips

        each_win = float(total) / sum(winners)

        payoffs = []
        for i, _ in enumerate(players):
            if winners[i] == 1:
                payoffs.append(each_win - players[i].in_chips)
            else:
                payoffs.append(float(-players[i].in_chips))

        return payoffs



import os
import csv

class Logger(object):
    ''' Logger saves the running results and helps make plots from the results
    '''

    def __init__(self, log_dir):
        ''' Initialize the labels, legend and paths of the plot and log file.

        Args:
            log_path (str): The path the log files
        '''
        self.log_dir = log_dir

    def __enter__(self):
        self.txt_path = os.path.join(self.log_dir, 'log.txt')
        self.csv_path = os.path.join(self.log_dir, 'performance.csv')
        self.fig_path = os.path.join(self.log_dir, 'fig.png')

        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)

        self.txt_file = open(self.txt_path, 'w')
        self.csv_file = open(self.csv_path, 'w')
        fieldnames = ['episode', 'reward']
        self.writer = csv.DictWriter(self.csv_file, fieldnames=fieldnames)
        self.writer.writeheader()

        return self

    def log(self, text):
        ''' Write the text to log file then print it.
        Args:
            text(string): text to log
        '''
        self.txt_file.write(text+'\n')
        self.txt_file.flush()
        print(text)

    def log_performance(self, episode, reward):
        ''' Log a point in the curve
        Args:
            episode (int): the episode of the current point
            reward (float): the reward of the current point
        '''
        self.writer.writerow({'episode': episode, 'reward': reward})
        print('')
        self.log('----------------------------------------')
        self.log('  episode      |  ' + str(episode))
        self.log('  reward       |  ' + str(reward))
        self.log('----------------------------------------')

    def __exit__(self, type, value, traceback):
        if self.txt_path is not None:
            self.txt_file.close()
        if self.csv_path is not None:
            self.csv_file.close()
        print('\nLogs saved in', self.log_dir)



from rlcard.games.limitholdem.utils import compare_hands
import numpy as np


class LimitHoldemJudger:
    """The Judger class for limit texas holdem"""

    def __init__(self, np_random):
        self.np_random = np_random

    def judge_game(self, players, hands):
        """
        Judge the winner of the game.

        Args:
            players (list): The list of players who play the game
            hands (list): The list of hands that from the players

        Returns:
            (list): Each entry of the list corresponds to one entry of the
        """
        # Convert the hands into card indexes
        hands = [[card.get_index() for card in hand] if hand is not None else None for hand in hands]
        
        in_chips = [p.in_chips for p in players]
        remaining = sum(in_chips)
        payoffs = [0] * len(hands)
        while remaining > 0:
            winners = compare_hands(hands)
            each_win = self.split_pots_among_players(in_chips, winners)
            
            for i in range(len(players)):
                if winners[i]:
                    remaining -= each_win[i]
                    payoffs[i] += each_win[i] - in_chips[i]
                    hands[i] = None
                    in_chips[i] = 0
                elif in_chips[i] > 0:
                    payoffs[i] += each_win[i] - in_chips[i]
                    in_chips[i] = each_win[i]
                    
        assert sum(payoffs) == 0
        return payoffs

    def split_pot_among_players(self, in_chips, winners):
        """
        Splits the next (side) pot among players.
        Function is called in loop by distribute_pots_among_players until all chips are allocated.

        Args:
            in_chips (list): List with number of chips bet not yet distributed for each player
            winners (list): List with 1 if the player is among winners else 0

        Returns:
            (list): Of how much chips each player get after this pot has been split and list of chips left to distribute
        """
        nb_winners_in_pot = sum((winners[i] and in_chips[i] > 0) for i in range(len(in_chips)))
        nb_players_in_pot = sum(in_chips[i] > 0 for i in range(len(in_chips)))
        if nb_winners_in_pot == 0 or nb_winners_in_pot == nb_players_in_pot:
            # no winner or all winners for this pot
            allocated = list(in_chips)  # we give back their chips to each players in this pot
            in_chips_after = len(in_chips) * [0]  # no more chips to distribute
        else:
            amount_in_pot_by_player = min(v for v in in_chips if v > 0)
            how_much_one_win, remaining = divmod(amount_in_pot_by_player * nb_players_in_pot, nb_winners_in_pot)
            '''
            In the event of a split pot that cannot be divided equally for every winner, the winner who is sitting 
            closest to the left of the dealer receives the remaining differential in chips cf 
            https://www.betclic.fr/poker/house-rules--play-safely--betclic-poker-cpok_rules to simplify and as this 
            case is very rare, we will give the remaining differential in chips to a random winner
            '''
            allocated = len(in_chips) * [0]
            in_chips_after = list(in_chips)
            for i in range(len(in_chips)):  # iterate on all players
                if in_chips[i] == 0:  # player not in pot
                    continue
                if winners[i]:
                    allocated[i] += how_much_one_win
                in_chips_after[i] -= amount_in_pot_by_player
            if remaining > 0:
                random_winning_player = self.np_random.choice(
                    [i for i in range(len(winners)) if winners[i] and in_chips[i] > 0])
                allocated[random_winning_player] += remaining
        assert sum(in_chips[i] - in_chips_after[i] for i in range(len(in_chips))) == sum(allocated)
        return allocated, in_chips_after

    def split_pots_among_players(self, in_chips_initial, winners):
        """
        Splits main pot and side pots among players (to handle special case of all-in players).

        Args:
            in_chips_initial (list): List with number of chips bet for each player
            winners (list): List with 1 if the player is among winners else 0

        Returns:
            (list): List of how much chips each player get back after all pots have been split
        """
        in_chips = list(in_chips_initial)
        assert len(in_chips) == len(winners)
        assert all(v == 0 or v == 1 for v in winners)
        assert sum(winners) >= 1  # there must be at least one winner
        allocated = np.zeros(len(in_chips), dtype=int)
        while any(v > 0 for v in in_chips):  # while there are still chips to allocate
            allocated_current_pot, in_chips = self.split_pot_among_players(in_chips, winners)
            allocated += allocated_current_pot  # element-wise addition
        assert all(chips >= 0 for chips in allocated)  # check that all players got a non negative amount of chips
        assert sum(in_chips_initial) == sum(allocated)  # check that all chips bet have been allocated
        return list(allocated)



from collections import Counter, OrderedDict
import numpy as np

from rlcard.envs import Env


class DoudizhuEnv(Env):
    ''' Doudizhu Environment
    '''

    def __init__(self, config):
        from rlcard.games.doudizhu.utils import ACTION_2_ID, ID_2_ACTION
        from rlcard.games.doudizhu.utils import cards2str, cards2str_with_suit
        from rlcard.games.doudizhu import Game
        self._cards2str = cards2str
        self._cards2str_with_suit = cards2str_with_suit
        self._ACTION_2_ID = ACTION_2_ID
        self._ID_2_ACTION = ID_2_ACTION
        
        self.name = 'doudizhu'
        self.game = Game()
        super().__init__(config)
        self.state_shape = [[790], [901], [901]]
        self.action_shape = [[54] for _ in range(self.num_players)]

    def _extract_state(self, state):
        ''' Encode state

        Args:
            state (dict): dict of original state
        '''
        current_hand = _cards2array(state['current_hand'])
        others_hand = _cards2array(state['others_hand'])

        last_action = ''
        if len(state['trace']) != 0:
            if state['trace'][-1][1] == 'pass':
                last_action = state['trace'][-2][1]
            else:
                last_action = state['trace'][-1][1]
        last_action = _cards2array(last_action)

        last_9_actions = _action_seq2array(_process_action_seq(state['trace']))

        if state['self'] == 0: # landlord
            landlord_up_played_cards = _cards2array(state['played_cards'][2])
            landlord_down_played_cards = _cards2array(state['played_cards'][1])
            landlord_up_num_cards_left = _get_one_hot_array(state['num_cards_left'][2], 17) 
            landlord_down_num_cards_left = _get_one_hot_array(state['num_cards_left'][1], 17)
            obs = np.concatenate((current_hand,
                                  others_hand,
                                  last_action,
                                  last_9_actions,
                                  landlord_up_played_cards,
                                  landlord_down_played_cards,
                                  landlord_up_num_cards_left,
                                  landlord_down_num_cards_left))
        else:
            landlord_played_cards = _cards2array(state['played_cards'][0])
            for i, action in reversed(state['trace']):
                if i == 0:
                    last_landlord_action = action
                    break
            last_landlord_action = _cards2array(last_landlord_action)
            landlord_num_cards_left = _get_one_hot_array(state['num_cards_left'][0], 20)

            teammate_id = 3 - state['self']
            teammate_played_cards = _cards2array(state['played_cards'][teammate_id])
            last_teammate_action = 'pass'
            for i, action in reversed(state['trace']):
                if i == teammate_id:
                    last_teammate_action = action
                    break
            last_teammate_action = _cards2array(last_teammate_action)
            teammate_num_cards_left = _get_one_hot_array(state['num_cards_left'][teammate_id], 17)
            obs = np.concatenate((current_hand,
                                  others_hand,
                                  last_action,
                                  last_9_actions,
                                  landlord_played_cards,
                                  teammate_played_cards,
                                  last_landlord_action,
                                  last_teammate_action,
                                  landlord_num_cards_left,
                                  teammate_num_cards_left))

        extracted_state = OrderedDict({'obs': obs, 'legal_actions': self._get_legal_actions()})
        extracted_state['raw_obs'] = state
        extracted_state['raw_legal_actions'] = [a for a in state['actions']]
        extracted_state['action_record'] = self.action_recorder
        return extracted_state
            
    def get_payoffs(self):
        ''' Get the payoffs of players. Must be implemented in the child class.

        Returns:
            payoffs (list): a list of payoffs for each player
        '''
        return self.game.judger.judge_payoffs(self.game.round.landlord_id, self.game.winner_id)

    def _decode_action(self, action_id):
        ''' Action id -> the action in the game. Must be implemented in the child class.

        Args:
            action_id (int): the id of the action

        Returns:
            action (string): the action that will be passed to the game engine.
        '''
        return self._ID_2_ACTION[action_id]

    def _get_legal_actions(self):
        ''' Get all legal actions for current state

        Returns:
            legal_actions (list): a list of legal actions' id
        '''
        legal_actions = self.game.state['actions']
        legal_actions = {self._ACTION_2_ID[action]: _cards2array(action) for action in legal_actions}
        return legal_actions

    def get_perfect_information(self):
        ''' Get the perfect information of the current state

        Returns:
            (dict): A dictionary of all the perfect information of the current state
        '''
        state = {}
        state['hand_cards_with_suit'] = [self._cards2str_with_suit(player.current_hand) for player in self.game.players]
        state['hand_cards'] = [self._cards2str(player.current_hand) for player in self.game.players]
        state['trace'] = self.game.state['trace']
        state['current_player'] = self.game.round.current_player
        state['legal_actions'] = self.game.state['actions']
        return state

    def get_action_feature(self, action):
        ''' For some environments such as DouDizhu, we can have action features

        Returns:
            (numpy.array): The action features
        '''
        return _cards2array(self._decode_action(action))

Card2Column = {'3': 0, '4': 1, '5': 2, '6': 3, '7': 4, '8': 5, '9': 6, 'T': 7,
               'J': 8, 'Q': 9, 'K': 10, 'A': 11, '2': 12}

NumOnes2Array = {0: np.array([0, 0, 0, 0]),
                 1: np.array([1, 0, 0, 0]),
                 2: np.array([1, 1, 0, 0]),
                 3: np.array([1, 1, 1, 0]),
                 4: np.array([1, 1, 1, 1])}

def _cards2array(cards):
    if cards == 'pass':
        return np.zeros(54, dtype=np.int8)

    matrix = np.zeros([4, 13], dtype=np.int8)
    jokers = np.zeros(2, dtype=np.int8)
    counter = Counter(cards)
    for card, num_times in counter.items():
        if card == 'B':
            jokers[0] = 1
        elif card == 'R':
            jokers[1] = 1
        else:
            matrix[:, Card2Column[card]] = NumOnes2Array[num_times]
    return np.concatenate((matrix.flatten('F'), jokers))

def _get_one_hot_array(num_left_cards, max_num_cards):
    one_hot = np.zeros(max_num_cards, dtype=np.int8)
    one_hot[num_left_cards - 1] = 1

    return one_hot

def _action_seq2array(action_seq_list):
    action_seq_array = np.zeros((len(action_seq_list), 54), np.int8)
    for row, cards in enumerate(action_seq_list):
        action_seq_array[row, :] = _cards2array(cards)
    action_seq_array = action_seq_array.flatten()
    return action_seq_array

def _process_action_seq(sequence, length=9):
    sequence = [action[1] for action in sequence[-length:]]
    if len(sequence) < length:
        empty_sequence = ['' for _ in range(length - len(sequence))]
        empty_sequence.extend(sequence)
        sequence = empty_sequence
    return sequence



from rlcard.utils import init_standard_deck
import numpy as np

class BlackjackDealer:

    def __init__(self, np_random, num_decks=1):
        ''' Initialize a Blackjack dealer class
        '''
        self.np_random = np_random
        self.num_decks = num_decks
        self.deck = init_standard_deck()
        if self.num_decks not in [0, 1]:  # 0 indicates infinite decks of cards
            self.deck = self.deck * self.num_decks  # copy m standard decks of cards
        self.shuffle()
        self.hand = []
        self.status = 'alive'
        self.score = 0

    def shuffle(self):
        ''' Shuffle the deck
        '''
        shuffle_deck = np.array(self.deck)
        self.np_random.shuffle(shuffle_deck)
        self.deck = list(shuffle_deck)

    def deal_card(self, player):
        ''' Distribute one card to the player

        Args:
            player_id (int): the target player's id
        '''
        idx = self.np_random.choice(len(self.deck))
        card = self.deck[idx]
        if self.num_decks != 0:  # If infinite decks, do not pop card from deck
            self.deck.pop(idx)
        # card = self.deck.pop()
        player.hand.append(card)



import unittest
import numpy as np
from rlcard.games.nolimitholdem.player import NolimitholdemPlayer as Player
from rlcard.games.base import Card
from rlcard.games.limitholdem.judger import LimitHoldemJudger as Judger
from rlcard.games.limitholdem.utils import Hand 


rand_state = np.random.RandomState()

class TestNolimitholdemGame(unittest.TestCase):

    def get_players(self, num_players=2):
        players = []
        
        for i in range(num_players):
            players.append(Player(i, 100 + 100*i, rand_state))
            players[i].bet(players[i].remained_chips) # All in
            
        return players
    
    def get_hands(self, player_hands, public_card):
        hands = []
        for hand in player_hands:
            hands.append(hand + public_card)
        return hands        
    
    def test_judge_with_4_players(self):

        '''
        suit_list = ['S', 'H', 'D', 'C']
        rank_list = ['A', '2', '3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K']
        '''
        players = self.get_players(4)
        
        
        public_card = [Card('S', 'A'), Card('S', 'K'), Card('S', 'Q'), Card('S', '2'), Card('S', '3')]
        hands = [[Card('S', 'J'), Card('S', 'T')],
                 [Card('S', '4'), Card('S', '5')], 
                 [Card('S', '9'), Card('C', 'T')], 
                 [Card('H', 'T'), Card('C', 'J')]]
        
        payoffs = Judger(rand_state).judge_game(players, self.get_hands(hands, public_card))
        self.assertEqual(payoffs, [300, 100, -100, -300])
        
        public_card = [Card('H', 'A'), Card('H', 'K'), Card('S', 'Q'), Card('S', 'T'), Card('S', '9')]
        
        hands = [[Card('S', 'A'), Card('H', '4')], 
                 [Card('D', 'A'), Card('H', '5')], 
                 [Card('D', 'K'), Card('H', '6')], 
                 [Card('S', 'K'), Card('H', '7')]]
        
        payoffs = Judger(rand_state).judge_game(players, self.get_hands(hands, public_card))
        self.assertEqual(payoffs, [100, 300, -200, -200])
        
    def test_judge_with_6_players(self):
        rand_state = np.random.RandomState()
        
        public_card = [Card('S', 'A'), Card('S', 'K'), Card('D', 'Q'), Card('D', 'T'), Card('C', '9')]
        players = self.get_players(6)
        
        hands = [[Card('C', 'A'), Card('H', '2')], 
                 [Card('D', 'A'), Card('H', '3')], 
                 [Card('C', 'K'), Card('C', '2')], 
                 [Card('D', 'K'), Card('C', '3')],
                 [Card('C', 'Q'), Card('S', '2')], 
                 [Card('D', 'Q'), Card('S', '3')]]

        payoffs = Judger(rand_state).judge_game(players, self.get_hands(hands, public_card))
        self.assertEqual(payoffs, [200, 600, -100, 100, -400, -400])


if __name__ == '__main__':
    unittest.main()



'''
    File name: gin_rummy/thinker.py
    Author: William Hale
    Date created: 3/28/2020
'''

from typing import List

from rlcard.games.base import Card

from . import melding
from . import utils


class Thinker(object):

    def __init__(self, hand: List[Card]):
        self.hand = hand

    # simple thinking
    def get_meld_piles_with_discard_card(self, discard_card: Card) -> List[List[Card]]:
        next_hand = self.hand + [discard_card]
        meld_clusters = melding.get_meld_clusters(hand=next_hand)
        best_deadwood_count = 999
        best_deadwoods = []  # type: List[List[Card]]
        best_meld_clusters = []  # type: List[List[List[Card]]]
        for meld_cluster in meld_clusters:
            meld_cards = [card for meld_pile in meld_cluster for card in meld_pile]
            deadwood = [card for card in next_hand if card not in meld_cards]
            deadwood_count = self._get_deadwood_count(deadwood=deadwood)
            if deadwood_count < best_deadwood_count:
                best_deadwood_count = deadwood_count
                best_deadwoods = [deadwood]
                best_meld_clusters = [meld_cluster]
            elif deadwood_count == best_deadwood_count:
                best_deadwoods.append(deadwood)
                best_meld_clusters.append(meld_cluster)
        want_discard_card = False
        for deadwood in best_deadwoods:
            if discard_card in deadwood:
                want_discard_card = False
                break
            else:
                want_discard_card = True
        result = []  # type: List[List[Card]]
        if want_discard_card:
            for meld_cluster in best_meld_clusters:
                for meld_pile in meld_cluster:
                    if discard_card in meld_pile:
                        result.append(meld_pile)
        return result

    @staticmethod
    def _get_deadwood_count(deadwood: List[Card]) -> int:
        deadwood_values = [utils.get_deadwood_value(card) for card in deadwood]
        return sum(deadwood_values)



'''
    File name: gin_rummy/dealer.py
    Author: William Hale
    Date created: 2/12/2020
'''

from .player import GinRummyPlayer
from .utils import utils as utils


class GinRummyDealer:
    ''' Initialize a GinRummy dealer class
    '''
    def __init__(self, np_random):
        ''' Empty discard_pile, set shuffled_deck, set stock_pile
        '''
        self.np_random = np_random
        self.discard_pile = []  # type: List[Card]
        self.shuffled_deck = utils.get_deck()  # keep a copy of the shuffled cards at start of new hand
        self.np_random.shuffle(self.shuffled_deck)
        self.stock_pile = self.shuffled_deck.copy()  # type: List[Card]

    def deal_cards(self, player: GinRummyPlayer, num: int):
        ''' Deal some cards from stock_pile to one player

        Args:
            player (GinRummyPlayer): The GinRummyPlayer object
            num (int): The number of cards to be dealt
        '''
        for _ in range(num):
            player.hand.append(self.stock_pile.pop())
        player.did_populate_hand()



from enum import Enum

import numpy as np
from copy import deepcopy
from rlcard.games.limitholdem import Game
from rlcard.games.limitholdem import PlayerStatus

from rlcard.games.nolimitholdem import Dealer
from rlcard.games.nolimitholdem import Player
from rlcard.games.nolimitholdem import Judger
from rlcard.games.nolimitholdem import Round, Action


class Stage(Enum):
    PREFLOP = 0
    FLOP = 1
    TURN = 2
    RIVER = 3
    END_HIDDEN = 4
    SHOWDOWN = 5


class NolimitholdemGame(Game):
    def __init__(self, allow_step_back=False, num_players=2):
        """Initialize the class no limit holdem Game"""
        super().__init__(allow_step_back, num_players)

        self.np_random = np.random.RandomState()

        # small blind and big blind
        self.small_blind = 1
        self.big_blind = 2 * self.small_blind

        # config players
        self.init_chips = [100] * num_players

        # If None, the dealer will be randomly chosen
        self.dealer_id = None

    def configure(self, game_config):
        """
        Specify some game specific parameters, such as number of players, initial chips, and dealer id.
        If dealer_id is None, he will be randomly chosen
        """
        self.num_players = game_config['game_num_players']
        # must have num_players length
        self.init_chips = [game_config['chips_for_each']] * game_config["game_num_players"]
        self.dealer_id = game_config['dealer_id']

    def init_game(self):
        """
        Initialize the game of not limit holdem

        This version supports two-player no limit texas holdem

        Returns:
            (tuple): Tuple containing:

                (dict): The first state of the game
                (int): Current player's id
        """
        if self.dealer_id is None:
            self.dealer_id = self.np_random.randint(0, self.num_players)

        # Initialize a dealer that can deal cards
        self.dealer = Dealer(self.np_random)

        # Initialize players to play the game
        self.players = [Player(i, self.init_chips[i], self.np_random) for i in range(self.num_players)]

        # Initialize a judger class which will decide who wins in the end
        self.judger = Judger(self.np_random)

        # Deal cards to each  player to prepare for the first round
        for i in range(2 * self.num_players):
            self.players[i % self.num_players].hand.append(self.dealer.deal_card())

        # Initialize public cards
        self.public_cards = []
        self.stage = Stage.PREFLOP

        # Big blind and small blind
        s = (self.dealer_id + 1) % self.num_players
        b = (self.dealer_id + 2) % self.num_players
        self.players[b].bet(chips=self.big_blind)
        self.players[s].bet(chips=self.small_blind)

        # The player next to the big blind plays the first
        self.game_pointer = (b + 1) % self.num_players

        # Initialize a bidding round, in the first round, the big blind and the small blind needs to
        # be passed to the round for processing.
        self.round = Round(self.num_players, self.big_blind, dealer=self.dealer, np_random=self.np_random)

        self.round.start_new_round(game_pointer=self.game_pointer, raised=[p.in_chips for p in self.players])

        # Count the round. There are 4 rounds in each game.
        self.round_counter = 0

        # Save the history for stepping back to the last state.
        self.history = []

        state = self.get_state(self.game_pointer)

        return state, self.game_pointer

    def get_legal_actions(self):
        """
        Return the legal actions for current player

        Returns:
            (list): A list of legal actions
        """
        return self.round.get_nolimit_legal_actions(players=self.players)

    def step(self, action):
        """
        Get the next state

        Args:
            action (str): a specific action. (call, raise, fold, or check)

        Returns:
            (tuple): Tuple containing:

                (dict): next player's state
                (int): next player id
        """

        if action not in self.get_legal_actions():
            print(action, self.get_legal_actions())
            print(self.get_state(self.game_pointer))
            raise Exception('Action not allowed')

        if self.allow_step_back:
            # First snapshot the current state
            r = deepcopy(self.round)
            b = self.game_pointer
            r_c = self.round_counter
            d = deepcopy(self.dealer)
            p = deepcopy(self.public_cards)
            ps = deepcopy(self.players)
            self.history.append((r, b, r_c, d, p, ps))

        # Then we proceed to the next round
        self.game_pointer = self.round.proceed_round(self.players, action)

        players_in_bypass = [1 if player.status in (PlayerStatus.FOLDED, PlayerStatus.ALLIN) else 0 for player in self.players]
        if self.num_players - sum(players_in_bypass) == 1:
            last_player = players_in_bypass.index(0)
            if self.round.raised[last_player] >= max(self.round.raised):
                # If the last player has put enough chips, he is also bypassed
                players_in_bypass[last_player] = 1

        # If a round is over, we deal more public cards
        if self.round.is_over():
            # Game pointer goes to the first player not in bypass after the dealer, if there is one
            self.game_pointer = (self.dealer_id + 1) % self.num_players
            if sum(players_in_bypass) < self.num_players:
                while players_in_bypass[self.game_pointer]:
                    self.game_pointer = (self.game_pointer + 1) % self.num_players

            # For the first round, we deal 3 cards
            if self.round_counter == 0:
                self.stage = Stage.FLOP
                self.public_cards.append(self.dealer.deal_card())
                self.public_cards.append(self.dealer.deal_card())
                self.public_cards.append(self.dealer.deal_card())
                if len(self.players) == np.sum(players_in_bypass):
                    self.round_counter += 1
            # For the following rounds, we deal only 1 card
            if self.round_counter == 1:
                self.stage = Stage.TURN
                self.public_cards.append(self.dealer.deal_card())
                if len(self.players) == np.sum(players_in_bypass):
                    self.round_counter += 1
            if self.round_counter == 2:
                self.stage = Stage.RIVER
                self.public_cards.append(self.dealer.deal_card())
                if len(self.players) == np.sum(players_in_bypass):
                    self.round_counter += 1

            self.round_counter += 1
            self.round.start_new_round(self.game_pointer)

        state = self.get_state(self.game_pointer)

        return state, self.game_pointer

    def get_state(self, player_id):
        """
        Return player's state

        Args:
            player_id (int): player id

        Returns:
            (dict): The state of the player
        """
        self.dealer.pot = np.sum([player.in_chips for player in self.players])

        chips = [self.players[i].in_chips for i in range(self.num_players)]
        legal_actions = self.get_legal_actions()
        state = self.players[player_id].get_state(self.public_cards, chips, legal_actions)
        state['stakes'] = [self.players[i].remained_chips for i in range(self.num_players)]
        state['current_player'] = self.game_pointer
        state['pot'] = self.dealer.pot
        state['stage'] = self.stage
        return state

    def step_back(self):
        """
        Return to the previous state of the game

        Returns:
            (bool): True if the game steps back successfully
        """
        if len(self.history) > 0:
            self.round, self.game_pointer, self.round_counter, self.dealer, self.public_cards, self.players = self.history.pop()
            self.stage = Stage(self.round_counter)
            return True
        return False

    def get_num_players(self):
        """
        Return the number of players in no limit texas holdem

        Returns:
            (int): The number of players in the game
        """
        return self.num_players

    def get_payoffs(self):
        """
        Return the payoffs of the game

        Returns:
            (list): Each entry corresponds to the payoff of one player
        """
        hands = [p.hand + self.public_cards if p.status in (PlayerStatus.ALIVE, PlayerStatus.ALLIN) else None for p in self.players]
        chips_payoffs = self.judger.judge_game(self.players, hands)
        return chips_payoffs

    @staticmethod
    def get_num_actions():
        """
        Return the number of applicable actions

        Returns:
            (int): The number of actions. There are 6 actions (call, raise_half_pot, raise_pot, all_in, check and fold)
        """
        return len(Action)



'''
    Project: Gui Gin Rummy
    File name: game_canvas_post_doing_action.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

from typing import List, Tuple

from rlcard.games.base import Card

import rlcard.games.gin_rummy.judge as judge
import rlcard.games.gin_rummy.utils.utils as gin_rummy_utils

from rlcard.games.gin_rummy.utils.melding import get_best_meld_clusters

from . import configurations
from . import status_messaging
from . import utils

from .configurations import DRAW_CARD_ACTION_ID, PICK_UP_DISCARD_ACTION_ID
from .configurations import DISCARD_ACTION_ID, KNOCK_ACTION_ID, GIN_ACTION_ID


class GameCanvasPostDoingAction(object):

    def __init__(self, game_canvas: 'GameCanvas'):
        self.game_canvas = game_canvas

    def post_do_get_card_action(self,
                                player_id: int,
                                drawn_card_item_id: int,
                                hit_item_id: int,
                                drawn_card_item_tag: int):
        game_canvas = self.game_canvas
        drawn_card_id = game_canvas.getter.get_card_id(card_item_id=drawn_card_item_id)
        drawn_card = game_canvas.card_items[drawn_card_id]
        if not drawn_card.is_face_up() and game_canvas.is_treating_as_human(player_id=player_id):
            drawn_card.set_card_id_face_up(face_up=True)
        selected_held_pile_item_ids = game_canvas.getter.get_selected_held_pile_item_ids(player_id=player_id)
        if selected_held_pile_item_ids:
            item_ids = selected_held_pile_item_ids + [drawn_card_item_id]
            utils.drop_item_ids(item_ids=item_ids, on_item_id=hit_item_id,
                                player_id=player_id, game_canvas=game_canvas)
            status_messaging.show_put_card_message(player_id=player_id, game_canvas=game_canvas)
            if drawn_card_item_tag == configurations.DISCARD_PILE_TAG:
                action = PICK_UP_DISCARD_ACTION_ID
            else:
                action = DRAW_CARD_ACTION_ID
            game_canvas.after_idle(game_canvas.game_canvas_updater.did_perform_actions, [action])
        else:
            to_location = game_canvas.bbox(hit_item_id)[:2]
            dx = game_canvas.held_pile_tab
            to_location = utils.translated_by(dx=dx, dy=0, location=to_location)

            def loop_completion():
                game_canvas.dtag(drawn_card_item_id, configurations.DRAWN_TAG)
                game_canvas.dtag(drawn_card_item_id, drawn_card_item_tag)
                game_canvas.addtag_withtag(game_canvas.held_pile_tags[player_id], drawn_card_item_id)
                if not game_canvas.is_treating_as_human(player_id=player_id):
                    utils.set_card_item_id_face_up(card_item_id=drawn_card_item_id,
                                                   face_up=False,
                                                   game_canvas=game_canvas)
                utils.held_pile_insert(card_item_id=drawn_card_item_id, above_hit_item_id=hit_item_id,
                                       player_id=player_id, game_canvas=game_canvas)
                if drawn_card_item_tag == configurations.DISCARD_PILE_TAG:
                    action = PICK_UP_DISCARD_ACTION_ID
                else:
                    action = DRAW_CARD_ACTION_ID
                game_canvas.after_idle(game_canvas.game_canvas_updater.did_perform_actions, [action])

            self._move_loop(drawn_card_item_id, to_location, completion=loop_completion)

    def post_do_discard_action(self, player_id: int, selected_held_pile_item_id: int):
        # The selected held_pile_item is discarded.
        # The remaining held_pile_items are fanned.
        game_canvas = self.game_canvas
        top_discard_pile_item_id = game_canvas.getter.get_top_discard_pile_item_id()
        if top_discard_pile_item_id is None:
            to_location = game_canvas.discard_pile_anchor
        else:
            dx = game_canvas.discard_pile_tab
            to_location = game_canvas.coords(top_discard_pile_item_id)
            to_location = utils.translated_by(dx=dx, dy=0, location=to_location)
        utils.set_card_item_id_face_up(card_item_id=selected_held_pile_item_id, face_up=True, game_canvas=game_canvas)

        if not game_canvas.query.is_human(player_id=player_id):  # turn card over immediately
            selected_card_id = game_canvas.getter.get_card_id(selected_held_pile_item_id)
            utils.set_card_id_face_up(selected_card_id, face_up=True, game_canvas=game_canvas)

        def loop_completion():
            game_canvas.dtag(selected_held_pile_item_id, configurations.SELECTED_TAG)
            game_canvas.dtag(selected_held_pile_item_id, configurations.JOGGED_TAG)
            game_canvas.dtag(selected_held_pile_item_id, game_canvas.held_pile_tags[player_id])
            game_canvas.addtag_withtag(configurations.DISCARD_PILE_TAG, selected_held_pile_item_id)
            utils.fan_held_pile(player_id=player_id, game_canvas=game_canvas)
            card_id = game_canvas.card_item_ids.index(selected_held_pile_item_id)
            action = DISCARD_ACTION_ID + card_id
            game_canvas.after_idle(game_canvas.game_canvas_updater.did_perform_actions, [action])

        self._move_loop(selected_held_pile_item_id, to_location, completion=loop_completion)

    def post_do_discard_card_drawn_from_stock_pile_action(self, top_stock_pile_item_id: int):
        game_canvas = self.game_canvas
        top_discard_pile_item_id = game_canvas.getter.get_top_discard_pile_item_id()

        dx = game_canvas.discard_pile_tab
        to_location = game_canvas.coords(top_discard_pile_item_id)
        to_location = utils.translated_by(dx=dx, dy=0, location=to_location)

        def loop_completion():
            game_canvas.dtag(top_stock_pile_item_id, configurations.DRAWN_TAG)
            game_canvas.dtag(top_stock_pile_item_id, configurations.STOCK_PILE_TAG)
            game_canvas.addtag_withtag(configurations.DISCARD_PILE_TAG, top_stock_pile_item_id)
            card_id = game_canvas.card_item_ids.index(top_stock_pile_item_id)
            actions = [DRAW_CARD_ACTION_ID, DISCARD_ACTION_ID + card_id]
            game_canvas.after_idle(game_canvas.game_canvas_updater.did_perform_actions, actions)

        self._move_loop(top_stock_pile_item_id, to_location, completion=loop_completion)

    def post_do_knock_action(self, selected_held_pile_item_id: int):
        # The selected held_pile_item is discarded.
        # The remaining held_pile_items are fanned.
        game_canvas = self.game_canvas
        game_canvas.going_out_button.place_forget()
        current_player_id = game_canvas.current_player_id
        top_discard_pile_item_id = game_canvas.getter.get_top_discard_pile_item_id()
        if top_discard_pile_item_id is None:
            to_location = game_canvas.discard_pile_anchor
        else:
            dx = game_canvas.discard_pile_tab
            to_location = game_canvas.coords(top_discard_pile_item_id)
            to_location = utils.translated_by(dx=dx, dy=0, location=to_location)
        utils.set_card_item_id_face_up(card_item_id=selected_held_pile_item_id, face_up=True, game_canvas=game_canvas)

        def loop_completion():
            game_canvas.dtag(selected_held_pile_item_id, configurations.SELECTED_TAG)
            game_canvas.dtag(selected_held_pile_item_id, configurations.JOGGED_TAG)
            game_canvas.dtag(selected_held_pile_item_id, game_canvas.held_pile_tags[current_player_id])
            game_canvas.addtag_withtag(configurations.DISCARD_PILE_TAG, selected_held_pile_item_id)
            utils.fan_held_pile(player_id=current_player_id, game_canvas=game_canvas)
            # show meld piles for both players
            self._show_meld_piles()
            # submit action to game_canvas_updater
            card_id = game_canvas.card_item_ids.index(selected_held_pile_item_id)
            action = KNOCK_ACTION_ID + card_id
            game_canvas.after_idle(game_canvas.game_canvas_updater.did_perform_actions, [action])

        self._move_loop(selected_held_pile_item_id, to_location, completion=loop_completion)

    def post_do_gin_action(self):
        game_canvas = self.game_canvas
        game_canvas.going_out_button.place_forget()
        current_player_id = game_canvas.current_player_id

        current_hand = game_canvas.getter.get_held_pile_cards(player_id=current_player_id)
        going_out_deadwood_count = self.game_canvas.game_canvas_updater.env.game.settings.going_out_deadwood_count
        _, gin_cards = judge.get_going_out_cards(hand=current_hand, going_out_deadwood_count=going_out_deadwood_count)
        card = gin_cards[0]
        card_id = gin_rummy_utils.get_card_id(card=card)
        card_item = game_canvas.card_items[card_id]
        selected_held_pile_item_id = card_item.item_id

        top_discard_pile_item_id = game_canvas.getter.get_top_discard_pile_item_id()
        if top_discard_pile_item_id is None:
            to_location = game_canvas.discard_pile_anchor
        else:
            dx = game_canvas.discard_pile_tab
            to_location = game_canvas.coords(top_discard_pile_item_id)
            to_location = utils.translated_by(dx=dx, dy=0, location=to_location)
        utils.set_card_item_id_face_up(card_item_id=selected_held_pile_item_id, face_up=True, game_canvas=game_canvas)

        def loop_completion():
            game_canvas.dtag(selected_held_pile_item_id, configurations.SELECTED_TAG)
            game_canvas.dtag(selected_held_pile_item_id, configurations.JOGGED_TAG)
            game_canvas.dtag(selected_held_pile_item_id, game_canvas.held_pile_tags[current_player_id])
            game_canvas.addtag_withtag(configurations.DISCARD_PILE_TAG, selected_held_pile_item_id)
            utils.fan_held_pile(player_id=current_player_id, game_canvas=game_canvas)
            # show meld piles for both players
            self._show_meld_piles()
            # submit action to game_canvas_updater
            action = GIN_ACTION_ID
            game_canvas.after_idle(game_canvas.game_canvas_updater.did_perform_actions, [action])

        self._move_loop(selected_held_pile_item_id, to_location, completion=loop_completion)

    def post_do_declare_dead_hand_action(self, player_id: int):
        game_canvas = self.game_canvas
        status_messaging.show_epilog_message_on_declare_dead_hand(game_canvas=game_canvas)
        game_canvas.going_out_button.place_forget()
        game_canvas.dead_hand_button.place_forget()
        # show meld piles for both players
        self._show_meld_piles()

    #   =========================================
    #   Private methods
    #   =========================================

    def _show_meld_piles(self):
        game_canvas = self.game_canvas
        current_player_id = game_canvas.current_player_id
        opponent_player_id = (current_player_id + 1) % 2
        utils.fan_held_pile(player_id=current_player_id, game_canvas=game_canvas)
        # do current_player_id melding
        best_meld_cluster = self._get_best_meld_cluster(player_id=current_player_id)
        self.put_down_meld_cluster(best_meld_cluster, player_id=current_player_id)
        # do opponent_player_id melding
        opponent_best_meld_cluster = self._get_best_meld_cluster(player_id=opponent_player_id)
        self.put_down_meld_cluster(opponent_best_meld_cluster, player_id=opponent_player_id)

    def _get_best_meld_cluster(self, player_id: int) -> List[List[Card]]:
        game_canvas = self.game_canvas
        hand = game_canvas.getter.get_held_pile_cards(player_id=player_id)
        best_meld_clusters = get_best_meld_clusters(hand=hand)
        best_meld_cluster = [] if not best_meld_clusters else best_meld_clusters[0]
        return best_meld_cluster

    def put_down_meld_cluster(self, meld_cluster, player_id: int):
        game_canvas = self.game_canvas
        card_width = game_canvas.card_width
        card_height = game_canvas.card_height
        player_pane = game_canvas.player_panes[player_id]
        y_tab = int(card_height * 0.15)
        anchor_x = int(card_width * 0.5)  # type: int
        anchor_y = int(game_canvas.coords(player_pane.item_id)[1]) + y_tab  # type: int
        for meld_pile in meld_cluster:
            self.put_down_meld_pile(meld_pile, anchor=(anchor_x, anchor_y), player_id=player_id)
            utils.fan_held_pile(player_id=player_id, game_canvas=game_canvas)
            anchor_x += len(meld_pile) * game_canvas.held_pile_tab
            anchor_y += y_tab
        held_pile_item_ids = game_canvas.getter.get_held_pile_item_ids(player_id=player_id)
        if not game_canvas.query.is_human(player_id):
            # sort deadwood cards of computer player
            held_pile_cards = game_canvas.getter.get_held_pile_cards(player_id=player_id)
            card_ids = [gin_rummy_utils.get_card_id(card) for card in held_pile_cards]
            sorted_card_ids = sorted(card_ids, reverse=True, key=utils.gin_rummy_sort_order_id)
            for sorted_card_id in sorted_card_ids:
                card_item_id = game_canvas.card_item_ids[sorted_card_id]
                game_canvas.tag_raise(card_item_id)
            utils.fan_held_pile(player_id, game_canvas=game_canvas)
            # face up deadwood cards of computer player
            for held_pile_item_id in held_pile_item_ids:
                utils.set_card_item_id_face_up(card_item_id=held_pile_item_id, face_up=True, game_canvas=game_canvas)

    def put_down_meld_pile(self, meld_pile: List[Card], anchor: Tuple[int, int], player_id: int):
        game_canvas = self.game_canvas
        held_pile_tag = game_canvas.held_pile_tags[player_id]
        x, y = anchor
        card_ids = [gin_rummy_utils.get_card_id(card=card) for card in meld_pile]
        sorted_card_ids = sorted(card_ids, reverse=True, key=utils.gin_rummy_sort_order_id)
        for sorted_card_id in sorted_card_ids:
            card_item_id = game_canvas.card_item_ids[sorted_card_id]
            game_canvas.tag_raise(card_item_id)
            game_canvas.dtag(card_item_id, held_pile_tag)
            utils.move_to(card_item_id, x, y, game_canvas)
            utils.set_card_item_id_face_up(card_item_id, face_up=True, game_canvas=game_canvas)
            x += game_canvas.held_pile_tab

    #   =========================================
    #   Utility Methods
    #   =========================================

    def _move_loop(self, item_id, to_location, index=0, dx=0, dy=0, completion=None):
        game_canvas = self.game_canvas
        step_size = 10
        time_in_milli_seconds = 10  # need to figure out the relationship between step_size and time_in_milli_seconds
        if index == 0:
            item_location = game_canvas.coords(item_id)
            dx = (to_location[0] - item_location[0]) / step_size
            dy = (to_location[1] - item_location[1]) / step_size
        if index == 1:  # Note: need to put 1 rather than 0. Should I just do tag_raise every iteration ???
            game_canvas.tag_raise(item_id)
            game_canvas.update_idletasks()  # Note: is this necessary ??? Probably not.
        game_canvas.move(item_id, dx, dy)
        index += 1
        if index < step_size:
            game_canvas.after(time_in_milli_seconds, self._move_loop, item_id, to_location, index, dx, dy, completion)
        else:
            game_canvas.coords(item_id, to_location)
            if completion:
                completion()



import numpy as np
from copy import copy

from rlcard.games.leducholdem import Dealer
from rlcard.games.leducholdem import Player
from rlcard.games.leducholdem import Judger
from rlcard.games.leducholdem import Round

from rlcard.games.limitholdem import Game

class LeducholdemGame(Game):

    def __init__(self, allow_step_back=False, num_players=2):
        ''' Initialize the class leducholdem Game
        '''
        self.allow_step_back = allow_step_back
        self.np_random = np.random.RandomState()
        ''' No big/small blind
        # Some configarations of the game
        # These arguments are fixed in Leduc Hold'em Game

        # Raise amount and allowed times
        self.raise_amount = 2
        self.allowed_raise_num = 2

        self.num_players = 2
        '''
        # Some configarations of the game
        # These arguments can be specified for creating new games

        # Small blind and big blind
        self.small_blind = 1
        self.big_blind = 2 * self.small_blind

        # Raise amount and allowed times
        self.raise_amount = self.big_blind
        self.allowed_raise_num = 2

        self.num_players = num_players

    def configure(self, game_config):
        ''' Specifiy some game specific parameters, such as number of players
        '''
        self.num_players = game_config['game_num_players']

    def init_game(self):
        ''' Initialilze the game of Limit Texas Hold'em

        This version supports two-player limit texas hold'em

        Returns:
            (tuple): Tuple containing:

                (dict): The first state of the game
                (int): Current player's id
        '''
        # Initilize a dealer that can deal cards
        self.dealer = Dealer(self.np_random)

        # Initilize two players to play the game
        self.players = [Player(i, self.np_random) for i in range(self.num_players)]

        # Initialize a judger class which will decide who wins in the end
        self.judger = Judger(self.np_random)

        # Prepare for the first round
        for i in range(self.num_players):
            self.players[i].hand = self.dealer.deal_card()
        # Randomly choose a small blind and a big blind
        s = self.np_random.randint(0, self.num_players)
        b = (s + 1) % self.num_players
        self.players[b].in_chips = self.big_blind
        self.players[s].in_chips = self.small_blind
        self.public_card = None
        # The player with small blind plays the first
        self.game_pointer = s

        # Initilize a bidding round, in the first round, the big blind and the small blind needs to
        # be passed to the round for processing.
        self.round = Round(raise_amount=self.raise_amount,
                           allowed_raise_num=self.allowed_raise_num,
                           num_players=self.num_players,
                           np_random=self.np_random)

        self.round.start_new_round(game_pointer=self.game_pointer, raised=[p.in_chips for p in self.players])

        # Count the round. There are 2 rounds in each game.
        self.round_counter = 0

        # Save the hisory for stepping back to the last state.
        self.history = []

        state = self.get_state(self.game_pointer)

        return state, self.game_pointer

    def step(self, action):
        ''' Get the next state

        Args:
            action (str): a specific action. (call, raise, fold, or check)

        Returns:
            (tuple): Tuple containing:

                (dict): next player's state
                (int): next plater's id
        '''
        if self.allow_step_back:
            # First snapshot the current state
            r = copy(self.round)
            r_raised = copy(self.round.raised)
            gp = self.game_pointer
            r_c = self.round_counter
            d_deck = copy(self.dealer.deck)
            p = copy(self.public_card)
            ps = [copy(self.players[i]) for i in range(self.num_players)]
            ps_hand = [copy(self.players[i].hand) for i in range(self.num_players)]
            self.history.append((r, r_raised, gp, r_c, d_deck, p, ps, ps_hand))

        # Then we proceed to the next round
        self.game_pointer = self.round.proceed_round(self.players, action)

        # If a round is over, we deal more public cards
        if self.round.is_over():
            # For the first round, we deal 1 card as public card. Double the raise amount for the second round
            if self.round_counter == 0:
                self.public_card = self.dealer.deal_card()
                self.round.raise_amount = 2 * self.raise_amount

            self.round_counter += 1
            self.round.start_new_round(self.game_pointer)

        state = self.get_state(self.game_pointer)

        return state, self.game_pointer

    def get_state(self, player):
        ''' Return player's state

        Args:
            player_id (int): player id

        Returns:
            (dict): The state of the player
        '''
        chips = [self.players[i].in_chips for i in range(self.num_players)]
        legal_actions = self.get_legal_actions()
        state = self.players[player].get_state(self.public_card, chips, legal_actions)
        state['current_player'] = self.game_pointer

        return state

    def is_over(self):
        ''' Check if the game is over

        Returns:
            (boolean): True if the game is over
        '''
        alive_players = [1 if p.status=='alive' else 0 for p in self.players]
        # If only one player is alive, the game is over.
        if sum(alive_players) == 1:
            return True

        # If all rounds are finshed
        if self.round_counter >= 2:
            return True
        return False

    def get_payoffs(self):
        ''' Return the payoffs of the game

        Returns:
            (list): Each entry corresponds to the payoff of one player
        '''
        chips_payoffs = self.judger.judge_game(self.players, self.public_card)
        payoffs = np.array(chips_payoffs) / (self.big_blind)
        return payoffs

    def step_back(self):
        ''' Return to the previous state of the game

        Returns:
            (bool): True if the game steps back successfully
        '''
        if len(self.history) > 0:
            self.round, r_raised, self.game_pointer, self.round_counter, d_deck, self.public_card, self.players, ps_hand = self.history.pop()
            self.round.raised = r_raised
            self.dealer.deck = d_deck
            for i, hand in enumerate(ps_hand):
                self.players[i].hand = hand
            return True
        return False



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import datetime
import csv
import json
import logging
import os
import time
from typing import Dict

import git


def gather_metadata() -> Dict:
    date_start = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
    # gathering git metadata
    try:
        repo = git.Repo(search_parent_directories=True)
        git_sha = repo.commit().hexsha
        git_data = dict(
            commit=git_sha,
            branch=repo.active_branch.name,
            is_dirty=repo.is_dirty(),
            path=repo.git_dir,
        )
    except git.InvalidGitRepositoryError:
        git_data = None
    # gathering slurm metadata
    if 'SLURM_JOB_ID' in os.environ:
        slurm_env_keys = [k for k in os.environ if k.startswith('SLURM')]
        slurm_data = {}
        for k in slurm_env_keys:
            d_key = k.replace('SLURM_', '').replace('SLURMD_', '').lower()
            slurm_data[d_key] = os.environ[k]
    else:
        slurm_data = None
    return dict(
        date_start=date_start,
        date_end=None,
        successful=False,
        git=git_data,
        slurm=slurm_data,
        env=os.environ.copy(),
    )


class FileWriter:
    def __init__(self,
                 xpid: str = None,
                 xp_args: dict = None,
                 rootdir: str = '~/palaas'):
        if not xpid:
            # make unique id
            xpid = '{proc}_{unixtime}'.format(
                proc=os.getpid(), unixtime=int(time.time()))
        self.xpid = xpid
        self._tick = 0

        # metadata gathering
        if xp_args is None:
            xp_args = {}
        self.metadata = gather_metadata()
        # we need to copy the args, otherwise when we close the file writer
        # (and rewrite the args) we might have non-serializable objects (or
        # other nasty stuff).
        self.metadata['args'] = copy.deepcopy(xp_args)
        self.metadata['xpid'] = self.xpid

        formatter = logging.Formatter('%(message)s')
        self._logger = logging.getLogger('palaas/out')

        # to stdout handler
        shandle = logging.StreamHandler()
        shandle.setFormatter(formatter)
        self._logger.addHandler(shandle)
        self._logger.setLevel(logging.INFO)

        rootdir = os.path.expandvars(os.path.expanduser(rootdir))
        # to file handler
        self.basepath = os.path.join(rootdir, self.xpid)

        if not os.path.exists(self.basepath):
            self._logger.info('Creating log directory: %s', self.basepath)
            os.makedirs(self.basepath, exist_ok=True)
        else:
            self._logger.info('Found log directory: %s', self.basepath)

        # NOTE: remove latest because it creates errors when running on slurm 
        # multiple jobs trying to write to latest but cannot find it 
        # Add 'latest' as symlink unless it exists and is no symlink.
        # symlink = os.path.join(rootdir, 'latest')
        # if os.path.islink(symlink):
        #     os.remove(symlink)
        # if not os.path.exists(symlink):
        #     os.symlink(self.basepath, symlink)
        #     self._logger.info('Symlinked log directory: %s', symlink)

        self.paths = dict(
            msg='{base}/out.log'.format(base=self.basepath),
            logs='{base}/logs.csv'.format(base=self.basepath),
            fields='{base}/fields.csv'.format(base=self.basepath),
            meta='{base}/meta.json'.format(base=self.basepath),
        )

        self._logger.info('Saving arguments to %s', self.paths['meta'])
        if os.path.exists(self.paths['meta']):
            self._logger.warning('Path to meta file already exists. '
                                 'Not overriding meta.')
        else:
            self._save_metadata()

        self._logger.info('Saving messages to %s', self.paths['msg'])
        if os.path.exists(self.paths['msg']):
            self._logger.warning('Path to message file already exists. '
                                 'New data will be appended.')

        fhandle = logging.FileHandler(self.paths['msg'])
        fhandle.setFormatter(formatter)
        self._logger.addHandler(fhandle)

        self._logger.info('Saving logs data to %s', self.paths['logs'])
        self._logger.info('Saving logs\' fields to %s', self.paths['fields'])
        if os.path.exists(self.paths['logs']):
            self._logger.warning('Path to log file already exists. '
                                 'New data will be appended.')
            with open(self.paths['fields'], 'r') as csvfile:
                reader = csv.reader(csvfile)
                self.fieldnames = list(reader)[0]
        else:
            self.fieldnames = ['_tick', '_time']

    def log(self, to_log: Dict, tick: int = None,
            verbose: bool = False) -> None:
        if tick is not None:
            raise NotImplementedError
        else:
            to_log['_tick'] = self._tick
            self._tick += 1
        to_log['_time'] = time.time()

        old_len = len(self.fieldnames)
        for k in to_log:
            if k not in self.fieldnames:
                self.fieldnames.append(k)
        if old_len != len(self.fieldnames):
            with open(self.paths['fields'], 'w') as csvfile:
                writer = csv.writer(csvfile)
                writer.writerow(self.fieldnames)
            self._logger.info('Updated log fields: %s', self.fieldnames)

        if to_log['_tick'] == 0:
            # print("\ncreating logs file ")
            with open(self.paths['logs'], 'a') as f:
                f.write('# %s\n' % ','.join(self.fieldnames))

        if verbose:
            self._logger.info('LOG | %s', ', '.join(
                ['{}: {}'.format(k, to_log[k]) for k in sorted(to_log)]))

        with open(self.paths['logs'], 'a') as f:
            writer = csv.DictWriter(f, fieldnames=self.fieldnames)
            writer.writerow(to_log)
            # print("\nadded to log file")

    def close(self, successful: bool = True) -> None:
        self.metadata['date_end'] = datetime.datetime.now().strftime(
            '%Y-%m-%d %H:%M:%S.%f')
        self.metadata['successful'] = successful
        self._save_metadata()

    def _save_metadata(self) -> None:
        with open(self.paths['meta'], 'w') as jsonfile:
            json.dump(self.metadata, jsonfile, indent=4, sort_keys=True)



from rlcard.games.limitholdem import Player


class NolimitholdemPlayer(Player):
    def __init__(self, player_id, init_chips, np_random):
        """
        Initialize a player.

        Args:
            player_id (int): The id of the player
            init_chips (int): The number of chips the player has initially
        """
        super().__init__(player_id, np_random)
        self.remained_chips = init_chips

    def bet(self, chips):
        quantity = chips if chips <= self.remained_chips else self.remained_chips
        self.in_chips += quantity
        self.remained_chips -= quantity



'''
    Project: Gui Gin Rummy
    File name: game_canvas.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_app import GameApp

import tkinter as tk

from typing import List, Dict

from ..gin_rummy_human_agent import HumanAgent
from ..gui_cards.card_image import CardImage, CardBackImage

from . import handling_tap_to_arrange_held_pile
from . import configurations
from . import handling_tap
from . import info_messaging

from .canvas_item import CardItem, CanvasItem
from .configurations import DECLARE_DEAD_HAND_ACTION_ID
from .game_canvas_debug import GameCanvasDebug
from .game_canvas_getter import GameCanvasGetter
from .game_canvas_post_doing_action import GameCanvasPostDoingAction
from .game_canvas_query import GameCanvasQuery
from .game_canvas_updater import GameCanvasUpdater
from .player_type import PlayerType

import rlcard.games.gin_rummy.utils.utils as gin_rummy_utils

from rlcard.games.gin_rummy.utils.action_event import KnockAction, GinAction
from rlcard.games.gin_rummy.utils.gin_rummy_error import GinRummyProgramError


class GameCanvas(tk.Canvas):

    def __init__(self, parent: tk.Frame, window_width, window_height, scale_factor, game_app: 'GameApp'):
        bottom_info_label_height = int(32 * scale_factor)
        canvas_width = window_width
        canvas_height = window_height - bottom_info_label_height
        super().__init__(master=parent, width=window_width,
                         height=window_height,
                         background=configurations.GAME_BACKGROUND_COLOR,
                         highlightthickness=0,
                         name="game_canvas")

        # keep list of relevant canvas items
        canvas_items = []  # type: [CanvasItem]

        # Make card images
        card_back_image = CardBackImage(scale_factor=scale_factor)
        card_images = {}  # type: Dict[(str, str)]  # card_images by rank by suit
        for card_id in range(52):
            card = gin_rummy_utils.card_from_card_id(card_id)
            card_image = CardImage(rank=card.rank, suit=card.suit, scale_factor=scale_factor)
            card_images[card.rank, card.suit] = card_image

        # Make card_item_ids by card_id
        card_item_ids = []  # type: List[int]
        for card_id in range(52):
            card = gin_rummy_utils.card_from_card_id(card_id)
            card_image = card_images[card.rank, card.suit]
            card_item_id = self.create_image((0, -9999), image=card_image, anchor="nw")
            self.itemconfigure(card_item_id, state=tk.HIDDEN)
            card_item_ids.append(card_item_id)

        # Make card_items by card_id
        card_items = []  # type: List[CardItem]
        for card_id in range(52):
            card = gin_rummy_utils.card_from_card_id(card_id)
            card_item_id = card_item_ids[card_id]
            card_image = card_images[card.rank, card.suit]
            card_item = CardItem(item_id=card_item_id, card_id=card_id, card_image=card_image, game_canvas=self)
            card_items.append(card_item)
            canvas_items.append(card_item)

        # Get card dimension
        card_width = card_images["A", "S"].width()
        card_height = card_images["A", "S"].height()

        # positioning information for stock_pile and discard_pile
        stock_pile_indent = card_width * 2
        stock_pile_tab = max(int(1 * scale_factor), 1)  # Note making int
        stock_pile_anchor = (stock_pile_indent, canvas_height / 2 - card_height / 2)
        discard_pile_anchor = (stock_pile_indent + card_width * 1.8, canvas_height / 2 - card_height / 2)

        # positioning information for held_piles
        held_pile_tab = max(int(card_width * 0.47), 1)  # Note making int
        max_held_pile_width = 10 * held_pile_tab + card_width
        north_held_pile_anchor = (canvas_width / 2 - max_held_pile_width / 2,
                                  stock_pile_anchor[1] - card_height * 1.5)
        south_held_pile_anchor = (canvas_width / 2 - max_held_pile_width / 2,
                                  stock_pile_anchor[1] + card_height * 1.5)

        # draw discard_pile_box
        discard_pile_box_left = discard_pile_anchor[0]
        discard_pile_box_top = discard_pile_anchor[1]
        discard_pile_box_right = discard_pile_box_left + card_width
        discard_pile_box_bottom = discard_pile_box_top + card_height
        discard_pile_box_item_id = self.create_rectangle(discard_pile_box_left,
                                                         discard_pile_box_top,
                                                         discard_pile_box_right,
                                                         discard_pile_box_bottom,
                                                         fill="gray")
        discard_pile_box_item = CanvasItem(item_id=discard_pile_box_item_id, game_canvas=self)
        canvas_items.append(discard_pile_box_item)

        # create player_panes
        player_panes = []
        for player_id in range(2):
            left = 0
            right = canvas_width
            top = 0 if player_id == 0 else stock_pile_anchor[1] + card_height
            bottom = stock_pile_anchor[1] if player_id == 0 else canvas_height
            player_pane_id = self.create_rectangle(left, top, right, bottom, width=0)
            player_pane = CanvasItem(item_id=player_pane_id, game_canvas=self)
            canvas_items.append(player_pane)
            player_panes.append(player_pane)

        # create two held_pile ghost card items
        held_pile_ghost_card_items = []
        player_held_pile_anchors = [north_held_pile_anchor, south_held_pile_anchor]
        held_pile_tags = [configurations.NORTH_HELD_PILE_TAG, configurations.SOUTH_HELD_PILE_TAG]
        for player_id in range(2):
            x, y = player_held_pile_anchors[player_id]
            x -= held_pile_tab
            ghost_card_item_id = self.create_rectangle(x, y, x + card_width, y + card_height, width=0, fill='')
            self.itemconfig(ghost_card_item_id, tag=held_pile_tags[player_id])
            ghost_card_item = CanvasItem(item_id=ghost_card_item_id, game_canvas=self)
            canvas_items.append(ghost_card_item)
            held_pile_ghost_card_items.append(ghost_card_item)

        # create bottom info_label
        info_label = tk.Label(self, text="", padx=10 * scale_factor, anchor="w")
        info_label.config(font=(None, int(bottom_info_label_height * 0.6)))
        info_label.place(x=0, y=canvas_height, width=window_width, height=bottom_info_label_height)

        # create going_out_button
        going_out_button = tk.Button(self, text="Knock", command=self.on_going_out)
        going_out_button.configure(font=(None, int(card_height * 0.15)))
        going_out_button_x = stock_pile_anchor[0] - card_width * 1.5
        going_out_button_y = stock_pile_anchor[1] + card_height / 2 - going_out_button.winfo_reqheight() / 2
        going_out_button.place(x=going_out_button_x, y=going_out_button_y)
        going_out_button_place = going_out_button.place_info()  # remember place
        going_out_button.place_forget()  # hide button

        # create dead_hand_button
        going_out_button_height = going_out_button.winfo_reqheight()
        dead_hand_button = tk.Button(self, text="Dead hand", command=self.on_dead_hand)
        dead_hand_button.configure(font=(None, int(card_height * 0.15)))
        dead_hand_button.place(x=going_out_button_x, y=going_out_button_y + going_out_button_height * 1.5)
        dead_hand_button_place = dead_hand_button.place_info()  # remember place
        dead_hand_button.place_forget()  # hide button

        # create score_pad widget
        cell_width = int(120 * scale_factor)
        cell_height = int(50 * scale_factor)
        cell_border_thickness = int(3 * scale_factor)
        cell_dy = int(5 * scale_factor)
        score_pad_widget = tk.Canvas(self, bg='white', highlightthickness=0)
        score_pad_widget.configure(width=2 * cell_width, height=2 * cell_height)
        score_pad_cells = []
        for row in range(2):
            for col in range(2):
                text = ""
                if row == 0:
                    text = ["Me", "You"][col]
                score_pad_cell = tk.Label(score_pad_widget, text=text, padx=0, pady=0, height=0, bg=None)
                score_pad_cell.config(width=7)  # Note: need to be fixed width to blank out
                score_pad_cell.config(font=("Times", int(cell_height * 0.6)))
                indent = 4 * scale_factor
                score_pad_cell.place(x=indent + col * cell_width, y=0 if row == 0 else cell_height)
                if row == 1:
                    score_pad_cells.append(score_pad_cell)
        indent_x = 10 * scale_factor
        indent_y = 5 * scale_factor
        score_pad_widget.create_line(indent_x, cell_height - indent_y, 2 * cell_width - indent_x,
                                     cell_height - indent_y, fill='black', width=cell_border_thickness)
        score_pad_widget.create_line(cell_width, cell_dy, cell_width, 2 * cell_height - cell_dy,
                                     fill='black', width=cell_border_thickness)
        score_pad_widget.pack()
        score_pad_widget.place(x=discard_pile_anchor[0] + 3 * card_width,
                               y=stock_pile_anchor[1] + card_height / 2 - cell_height)

        # create info_message_label
        info_message_label = tk.Label(self, anchor='nw', justify='left')
        info_message_label.configure(wraplength=185 * scale_factor)  # in pixels
        info_message_label.configure(width=0, height=0)  # dimensions in text units; wraplength determines width
        info_message_label.place(x=10 * scale_factor, y=stock_pile_anchor[1])

        # set left mouse button action for target canvas
        self.bind("<Button-1>", handling_tap.on_game_canvas_tap)
        self.bind("<Shift-Button-1>", handling_tap_to_arrange_held_pile.on_tap_to_arrange_held_pile)
        self.bind("<Button-2>", handling_tap_to_arrange_held_pile.on_tap_to_arrange_held_pile)

        # dictionary of item_id to canvas_item with that item_id
        canvas_item_by_item_id = {}
        for canvas_item in canvas_items:
            canvas_item_by_item_id[canvas_item.item_id] = canvas_item

        # keep as instance variables
        self.scale_factor = scale_factor
        self.held_pile_tags = held_pile_tags
        self.dealer_id = 0  # type: int  # Note: has no effect; will be set later to correct value
        self.canvas_items = canvas_items
        self.card_back_image = card_back_image
        self.card_images = card_images
        self.card_item_ids = card_item_ids
        self.card_items = card_items
        self.card_width = card_width
        self.card_height = card_height
        self.stock_pile_anchor = stock_pile_anchor
        self.stock_pile_tab = stock_pile_tab
        self.discard_pile_anchor = discard_pile_anchor
        self.discard_pile_tab = stock_pile_tab
        self.discard_pile_box_item = discard_pile_box_item
        self.player_panes = player_panes
        self.player_held_pile_anchors = player_held_pile_anchors
        self.held_pile_tab = held_pile_tab
        self.held_pile_ghost_card_items = held_pile_ghost_card_items
        self.info_label = info_label
        self.going_out_button_place = going_out_button_place
        self.going_out_button = going_out_button
        self.dead_hand_button_place = dead_hand_button_place
        self.dead_hand_button = dead_hand_button
        self.score_pad_cells = score_pad_cells
        self.info_message_label = info_message_label
        self.game_app = game_app
        self.canvas_item_by_item_id = canvas_item_by_item_id
        # delegates
        self.debug = GameCanvasDebug(game_canvas=self)
        self.getter = GameCanvasGetter(game_canvas=self)
        self.post_doing_action = GameCanvasPostDoingAction(game_canvas=self)
        self.query = GameCanvasQuery(game_canvas=self)

        self.pack()

        info_message_background = configurations.GAME_BACKGROUND_COLOR
        self._configure_info_message_label_colors(background_color=info_message_background)
        self.game_canvas_updater = GameCanvasUpdater(game_canvas=self)

    @property
    def current_player_id(self) -> int or None:  # convenience property to briefly get current_player_id
        return self.getter.get_current_player_id()

    @property
    def player_types(self) -> List[PlayerType]:
        result = []  # type: List[PlayerType]
        for agent in self.game_canvas_updater.env.agents:
            if isinstance(agent, HumanAgent):
                result.append(PlayerType.human_player)
            else:
                result.append(PlayerType.computer_player)
        return result

    def is_treating_as_human(self, player_id) -> bool:
        result = False
        player_types = self.player_types
        if player_types[player_id] == PlayerType.human_player:
            result = True
        elif player_id == 1 and player_types[0] == PlayerType.computer_player:
            result = True  # treat south player as human if both players are computer
        return result

    def update_configurations(self):
        self.update_configuration_game_background_color(background_color=configurations.GAME_BACKGROUND_COLOR)
        if not configurations.IS_SHOW_TIPS:
            info_messaging.blank_info_message_label(game_canvas=self)

    def update_configuration_game_background_color(self, background_color):
        self.configure(background=background_color)
        self._configure_info_message_label_colors(background_color=background_color)

    #   =========================================
    #   Command Methods
    #   =========================================

    def on_dead_hand(self):
        current_player_id = self.getter.get_current_player_id()
        self.post_doing_action.post_do_declare_dead_hand_action(player_id=current_player_id)
        # submit action to game_canvas_updater
        self.after_idle(self.game_canvas_updater.did_perform_actions, [DECLARE_DEAD_HAND_ACTION_ID])

    def on_going_out(self):
        if self.query.is_game_over():
            raise GinRummyProgramError("Cannot go out if game is over.")
        current_player_id = self.current_player_id
        selected_held_pile_item_ids = self.getter.get_selected_held_pile_item_ids(player_id=current_player_id)
        legal_actions = self.getter.get_legal_actions(player_id=current_player_id)
        gin_action = GinAction()
        if gin_action in legal_actions:
            self.post_doing_action.post_do_gin_action()
        elif len(selected_held_pile_item_ids) == 1:
            selected_held_pile_item_id = selected_held_pile_item_ids[0]
            card_id = self.card_item_ids.index(selected_held_pile_item_id)
            card = gin_rummy_utils.get_card(card_id=card_id)
            knock_action = KnockAction(card=card)
            if knock_action in legal_actions:
                self.post_doing_action.post_do_knock_action(selected_held_pile_item_id)
            else:
                pass  # FIXME: need message
        else:
            pass  # FIXME: need message

    # Private methods

    def _configure_info_message_label_colors(self, background_color):
        info_message_label = self.info_message_label
        info_message_background = background_color
        if info_message_background[1:3] == '00' and info_message_background[5:7] == '00':
            info_message_label.configure(foreground='yellow', background=info_message_background)
            info_message_label.configure(borderwidth=0, relief='flat')
        else:
            info_message_label.configure(foreground='black', background='white')
            info_message_label.configure(borderwidth=1 * self.scale_factor, relief='solid')



from collections import OrderedDict

from .model import DMCAgent
from rlcard.utils.pettingzoo_utils import wrap_state


class DMCAgentPettingZoo(DMCAgent):
    def step(self, state):
        return super().step(wrap_state(state))

    def eval_step(self, state):
        return super().eval_step(wrap_state(state))

    def feed(self, ts):
        state, action, reward, next_state, done = tuple(ts)
        state = wrap_state(state)
        next_state = wrap_state(next_state)
        ts = (state, action, reward, next_state, done)
        return super().feed(ts)


class DMCModelPettingZoo:
    def __init__(
        self,
        env,
        mlp_layers=[512,512,512,512,512],
        exp_epsilon=0.01,
        device="0"
    ):
        self.agents = OrderedDict()
        for agent_name in env.agents:
            agent = DMCAgentPettingZoo(
                env.observation_space(agent_name)["observation"].shape,
                (env.action_space(agent_name).n,),
                mlp_layers,
                exp_epsilon,
                device,
            )
            self.agents[agent_name] = agent

    def share_memory(self):
        for agent in self.agents.values():
            agent.share_memory()

    def eval(self):
        for agent in self.agents.values():
            agent.eval()

    def parameters(self, index):
        return list(self.agents.values())[index].parameters()

    def get_agent(self, index):
        return list(self.agents.values())[index]

    def get_agents(self):
        return list(self.agents.values())



'''
    File name: envs/gin_rummy.py
    Author: William Hale
    Date created: 2/12/2020
'''
import numpy as np
from collections import OrderedDict

from rlcard.envs import Env

class GinRummyEnv(Env):
    ''' GinRummy Environment
    '''
    def __init__(self, config):
        from rlcard.games.gin_rummy.utils.move import ScoreSouthMove
        from rlcard.games.gin_rummy.utils import utils
        from rlcard.games.gin_rummy import Game
        self._ScoreSouthMove = ScoreSouthMove
        self._utils = utils

        self.name = 'gin-rummy'
        self.game = Game()
        super().__init__(config=config)
        self.state_shape = [[5, 52] for _ in range(self.num_players)]
        self.action_shape = [None for _ in range(self.num_players)]

    def _extract_state(self, state):  # 200213 don't use state ???
        ''' Encode state

        Args:
            state (dict): dict of original state

        Returns:
            numpy array: 5 * 52 array
                         5 : current hand (1 if card in hand else 0)
                             top_discard (1 if card is top discard else 0)
                             dead_cards (1 for discards except for top_discard else 0)
                             opponent known cards (likewise)
                             unknown cards (likewise)  # is this needed ??? 200213
        '''
        if self.game.is_over():
            obs = np.array([self._utils.encode_cards([]) for _ in range(5)])
            extracted_state = {'obs': obs, 'legal_actions': self._get_legal_actions()}
            extracted_state['raw_legal_actions'] = list(self._get_legal_actions().keys())
            extracted_state['raw_obs'] = obs
        else:
            discard_pile = self.game.round.dealer.discard_pile
            stock_pile = self.game.round.dealer.stock_pile
            top_discard = [] if not discard_pile else [discard_pile[-1]]
            dead_cards = discard_pile[:-1]
            current_player = self.game.get_current_player()
            opponent = self.game.round.players[(current_player.player_id + 1) % 2]
            known_cards = opponent.known_cards
            unknown_cards = stock_pile + [card for card in opponent.hand if card not in known_cards]
            hand_rep = self._utils.encode_cards(current_player.hand)
            top_discard_rep = self._utils.encode_cards(top_discard)
            dead_cards_rep = self._utils.encode_cards(dead_cards)
            known_cards_rep = self._utils.encode_cards(known_cards)
            unknown_cards_rep = self._utils.encode_cards(unknown_cards)
            rep = [hand_rep, top_discard_rep, dead_cards_rep, known_cards_rep, unknown_cards_rep]
            obs = np.array(rep)
            extracted_state = {'obs': obs, 'legal_actions': self._get_legal_actions(), 'raw_legal_actions': list(self._get_legal_actions().keys())}
            extracted_state['raw_obs'] = obs
        return extracted_state

    def get_payoffs(self):
        ''' Get the payoffs of players. Must be implemented in the child class.

        Returns:
            payoffs (list): a list of payoffs for each player
        '''
        # determine whether game completed all moves
        is_game_complete = False
        if self.game.round:
            move_sheet = self.game.round.move_sheet
            if move_sheet and isinstance(move_sheet[-1], self._ScoreSouthMove):
                is_game_complete = True
        payoffs = [0, 0] if not is_game_complete else self.game.judge.scorer.get_payoffs(game=self.game)
        return np.array(payoffs)

    def _decode_action(self, action_id):  # FIXME 200213 should return str
        ''' Action id -> the action in the game. Must be implemented in the child class.

        Args:
            action_id (int): the id of the action

        Returns:
            action (ActionEvent): the action that will be passed to the game engine.
        '''
        return self.game.decode_action(action_id=action_id)

    def _get_legal_actions(self):
        ''' Get all legal actions for current state

        Returns:
            legal_actions (list): a list of legal actions' id
        '''
        legal_actions = self.game.judge.get_legal_actions()
        legal_actions_ids = {action_event.action_id: None for action_event in legal_actions}
        return OrderedDict(legal_actions_ids)



import numpy as np
from collections import OrderedDict

from rlcard.envs import Env
from rlcard.games.blackjack import Game

DEFAULT_GAME_CONFIG = {
        'game_num_players': 1,
        'game_num_decks': 1
        }

class BlackjackEnv(Env):
    ''' Blackjack Environment
    '''

    def __init__(self, config):
        ''' Initialize the Blackjack environment
        '''
        self.name = 'blackjack'
        self.default_game_config = DEFAULT_GAME_CONFIG
        self.game = Game()
        super().__init__(config)
        self.actions = ['hit', 'stand']
        self.state_shape = [[2] for _ in range(self.num_players)]
        self.action_shape = [None for _ in range(self.num_players)]

    def _get_legal_actions(self):
        ''' Get all leagal actions

        Returns:
            encoded_action_list (list): return encoded legal action list (from str to int)
        '''
        encoded_action_list = []
        for i in range(len(self.actions)):
            encoded_action_list.append(i)
        return encoded_action_list

    def _extract_state(self, state):
        ''' Extract the state representation from state dictionary for agent

        Args:
            state (dict): Original state from the game

        Returns:
            observation (list): combine the player's score and dealer's observable score for observation
        '''
        cards = state['state']
        my_cards = cards[0]
        dealer_cards = cards[1]

        my_score = get_score(my_cards)
        dealer_score = get_score(dealer_cards)
        obs = np.array([my_score, dealer_score])

        legal_actions = OrderedDict({i: None for i in range(len(self.actions))})
        extracted_state = {'obs': obs, 'legal_actions': legal_actions}
        extracted_state['raw_obs'] = state
        extracted_state['raw_legal_actions'] = [a for a in self.actions]
        extracted_state['action_record'] = self.action_recorder
        return extracted_state

    def get_payoffs(self):
        ''' Get the payoff of a game

        Returns:
           payoffs (list): list of payoffs
        '''
        payoffs = []

        for i in range(self.num_players):
            if self.game.winner['player' + str(i)] == 2:
                payoffs.append(1)  # Dealer bust or player get higher score than dealer
            elif self.game.winner['player' + str(i)] == 1:
                payoffs.append(0)  # Dealer and player tie
            else:
                payoffs.append(-1)  # Player bust or Dealer get higher score than player

        return np.array(payoffs)


    def _decode_action(self, action_id):
        ''' Decode the action for applying to the game

        Args:
            action id (int): action id

        Returns:
            action (str): action for the game
        '''
        return self.actions[action_id]

rank2score = {"A":11, "2":2, "3":3, "4":4, "5":5, "6":6, "7":7, "8":8, "9":9, "T":10, "J":10, "Q":10, "K":10}
def get_score(hand):
    score = 0
    count_a = 0
    for card in hand:
        score += rank2score[card[1:]]
        if card[1] == 'A':
            count_a += 1
    while score > 21 and count_a > 0:
        count_a -= 1
        score -= 10
    return score



import numpy as np

from rlcard.games.base import Card

def set_seed(seed):
    if seed is not None:
        import subprocess
        import sys

        reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'])
        installed_packages = [r.decode().split('==')[0] for r in reqs.split()]
        if 'torch' in installed_packages:
            import torch
            torch.backends.cudnn.deterministic = True
            torch.manual_seed(seed)
        np.random.seed(seed)
        import random
        random.seed(seed)

def get_device():
    import torch
    if torch.backends.mps.is_available():
        device = torch.device("mps:0")
        print("--> Running on the GPU")
    elif torch.cuda.is_available():
        device = torch.device("cuda:0")
        print("--> Running on the GPU")
    else:
        device = torch.device("cpu")
        print("--> Running on the CPU")

    return device    

def init_standard_deck():
    ''' Initialize a standard deck of 52 cards

    Returns:
        (list): A list of Card object
    '''
    suit_list = ['S', 'H', 'D', 'C']
    rank_list = ['A', '2', '3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K']
    res = [Card(suit, rank) for suit in suit_list for rank in rank_list]
    return res

def init_54_deck():
    ''' Initialize a standard deck of 52 cards, BJ and RJ

    Returns:
        (list): Alist of Card object
    '''
    suit_list = ['S', 'H', 'D', 'C']
    rank_list = ['A', '2', '3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K']
    res = [Card(suit, rank) for suit in suit_list for rank in rank_list]
    res.append(Card('BJ', ''))
    res.append(Card('RJ', ''))
    return res

def rank2int(rank):
    ''' Get the coresponding number of a rank.

    Args:
        rank(str): rank stored in Card object

    Returns:
        (int): the number corresponding to the rank

    Note:
        1. If the input rank is an empty string, the function will return -1.
        2. If the input rank is not valid, the function will return None.
    '''
    if rank == '':
        return -1
    elif rank.isdigit():
        if int(rank) >= 2 and int(rank) <= 10:
            return int(rank)
        else:
            return None
    elif rank == 'A':
        return 14
    elif rank == 'T':
        return 10
    elif rank == 'J':
        return 11
    elif rank == 'Q':
        return 12
    elif rank == 'K':
        return 13
    return None

def elegent_form(card):
    ''' Get a elegent form of a card string

    Args:
        card (string): A card string

    Returns:
        elegent_card (string): A nice form of card
    '''
    suits = {'S': '♠', 'H': '♥', 'D': '♦', 'C': '♣','s': '♠', 'h': '♥', 'd': '♦', 'c': '♣' }
    rank = '10' if card[1] == 'T' else card[1]

    return suits[card[0]] + rank

def print_card(cards):
    ''' Nicely print a card or list of cards

    Args:
        card (string or list): The card(s) to be printed
    '''
    if cards is None:
        cards = [None]
    if isinstance(cards, str):
        cards = [cards]

    lines = [[] for _ in range(9)]

    for card in cards:
        if card is None:
            lines[0].append('┌─────────┐')
            lines[1].append('│░░░░░░░░░│')
            lines[2].append('│░░░░░░░░░│')
            lines[3].append('│░░░░░░░░░│')
            lines[4].append('│░░░░░░░░░│')
            lines[5].append('│░░░░░░░░░│')
            lines[6].append('│░░░░░░░░░│')
            lines[7].append('│░░░░░░░░░│')
            lines[8].append('└─────────┘')
        else:
            if isinstance(card, Card):
                elegent_card = elegent_form(card.suit + card.rank)
            else:
                elegent_card = elegent_form(card)
            suit = elegent_card[0]
            rank = elegent_card[1]
            if len(elegent_card) == 3:
                space = elegent_card[2]
            else:
                space = ' '

            lines[0].append('┌─────────┐')
            lines[1].append('│{}{}       │'.format(rank, space))
            lines[2].append('│         │')
            lines[3].append('│         │')
            lines[4].append('│    {}    │'.format(suit))
            lines[5].append('│         │')
            lines[6].append('│         │')
            lines[7].append('│       {}{}│'.format(space, rank))
            lines[8].append('└─────────┘')

    for line in lines:
        print ('   '.join(line))

def reorganize(trajectories, payoffs):
    ''' Reorganize the trajectory to make it RL friendly

    Args:
        trajectory (list): A list of trajectories
        payoffs (list): A list of payoffs for the players. Each entry corresponds to one player

    Returns:
        (list): A new trajectories that can be fed into RL algorithms.

    '''
    num_players = len(trajectories)
    new_trajectories = [[] for _ in range(num_players)]

    for player in range(num_players):
        for i in range(0, len(trajectories[player])-2, 2):
            if i ==len(trajectories[player])-3:
                reward = payoffs[player]
                done =True
            else:
                reward, done = 0, False
            transition = trajectories[player][i:i+3].copy()
            transition.insert(2, reward)
            transition.append(done)

            new_trajectories[player].append(transition)
    return new_trajectories

def remove_illegal(action_probs, legal_actions):
    ''' Remove illegal actions and normalize the
        probability vector

    Args:
        action_probs (numpy.array): A 1 dimention numpy array.
        legal_actions (list): A list of indices of legal actions.

    Returns:
        probd (numpy.array): A normalized vector without legal actions.
    '''
    probs = np.zeros(action_probs.shape[0])
    probs[legal_actions] = action_probs[legal_actions]
    if np.sum(probs) == 0:
        probs[legal_actions] = 1 / len(legal_actions)
    else:
        probs /= sum(probs)
    return probs

def tournament(env, num):
    ''' Evaluate he performance of the agents in the environment

    Args:
        env (Env class): The environment to be evaluated.
        num (int): The number of games to play.

    Returns:
        A list of avrage payoffs for each player
    '''
    payoffs = [0 for _ in range(env.num_players)]
    counter = 0
    while counter < num:
        _, _payoffs = env.run(is_training=False)
        if isinstance(_payoffs, list):
            for _p in _payoffs:
                for i, _ in enumerate(payoffs):
                    payoffs[i] += _p[i]
                counter += 1
        else:
            for i, _ in enumerate(payoffs):
                payoffs[i] += _payoffs[i]
            counter += 1
    for i, _ in enumerate(payoffs):
        payoffs[i] /= counter
    return payoffs

def plot_curve(csv_path, save_path, algorithm):
    ''' Read data from csv file and plot the results
    '''
    import os
    import csv
    import matplotlib.pyplot as plt
    with open(csv_path) as csvfile:
        reader = csv.DictReader(csvfile)
        xs = []
        ys = []
        for row in reader:
            xs.append(int(row['episode']))
            ys.append(float(row['reward']))
        fig, ax = plt.subplots()
        ax.plot(xs, ys, label=algorithm)
        ax.set(xlabel='episode', ylabel='reward')
        ax.legend()
        ax.grid()

        save_dir = os.path.dirname(save_path)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        fig.savefig(save_path)




import rlcard
from rlcard.agents.random_agent import RandomAgent
import random
import numpy as np

def hash_obsevation(obs):
    try:
        val = hash(obs.tobytes())
        return val
    except AttributeError:
        try:
            return hash(obs)
        except TypeError:
            warnings.warn("Observation not an int or an Numpy array")
            return 0

def rand_iter(n):
    for x in range(n+1):
        random.randint(0, 1000)
        np.random.normal(size=100)

def gather_observations(env, actions, num_rand_steps):
    rand_iter(num_rand_steps)
    state, player_id = env.reset()
    rand_iter(num_rand_steps)

    action_idx = 0
    observations = []
    while not env.is_over() and action_idx < len(actions):
        # Agent plays
        rand_iter(num_rand_steps)
        legals = list(state['legal_actions'].keys())
        action = legals[actions[action_idx]%len(legals)]
        # Environment steps
        next_state, next_player_id = env.step(action)
        # Set the state and player
        state = next_state
        player_id = next_player_id

        action_idx += 1
        # Save state.
        if not env.game.is_over():
            observations.append(state)

    return observations

def is_deterministic(env_name):
    env = rlcard.make(env_name)

    NUM_STEPS = 25

    actions = [random.randrange(env.game.get_num_actions()) for _ in range(NUM_STEPS)]
    base_seed = 12941
    hashes = []
    for rand_iters in range(2):
        env = rlcard.make(env_name,config={'seed':base_seed})

        hashes.append(hash(tuple([hash_obsevation(obs['obs']) for obs in gather_observations(env,actions,rand_iters)])))

    return hashes[0] == hashes[1]



# Copyright 2019 Matthew Judell. All rights reserved.
# Copyright 2019 DATA Lab at Texas A&M University. All rights reserved.
# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' Neural Fictitious Self-Play (NFSP) agent implemented in TensorFlow.

See the paper https://arxiv.org/abs/1603.01121 for more details.
'''

import os
import random
import collections
import enum
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from rlcard.agents.dqn_agent import DQNAgent
from rlcard.utils.utils import remove_illegal

Transition = collections.namedtuple('Transition', 'info_state action_probs')

class NFSPAgent(object):
    ''' An approximate clone of rlcard.agents.nfsp_agent that uses
    pytorch instead of tensorflow.  Note that this implementation
    differs from Henrich and Silver (2016) in that the supervised
    training minimizes cross-entropy with respect to the stored
    action probabilities rather than the realized actions.
    '''

    def __init__(self,
                 num_actions=4,
                 state_shape=None,
                 hidden_layers_sizes=None,
                 reservoir_buffer_capacity=20000,
                 anticipatory_param=0.1,
                 batch_size=256,
                 train_every=1,
                 rl_learning_rate=0.1,
                 sl_learning_rate=0.005,
                 min_buffer_size_to_learn=100,
                 q_replay_memory_size=20000,
                 q_replay_memory_init_size=100,
                 q_update_target_estimator_every=1000,
                 q_discount_factor=0.99,
                 q_epsilon_start=0.06,
                 q_epsilon_end=0,
                 q_epsilon_decay_steps=int(1e6),
                 q_batch_size=32,
                 q_train_every=1,
                 q_mlp_layers=None,
                 evaluate_with='average_policy',
                 device=None,
                 save_path=None,
                 save_every=float('inf')):
        ''' Initialize the NFSP agent.

        Args:
            num_actions (int): The number of actions.
            state_shape (list): The shape of the state space.
            hidden_layers_sizes (list): The hidden layers sizes for the layers of
              the average policy.
            reservoir_buffer_capacity (int): The size of the buffer for average policy.
            anticipatory_param (float): The hyper-parameter that balances rl/avarage policy.
            batch_size (int): The batch_size for training average policy.
            train_every (int): Train the SL policy every X steps.
            rl_learning_rate (float): The learning rate of the RL agent.
            sl_learning_rate (float): the learning rate of the average policy.
            min_buffer_size_to_learn (int): The minimum buffer size to learn for average policy.
            q_replay_memory_size (int): The memory size of inner DQN agent.
            q_replay_memory_init_size (int): The initial memory size of inner DQN agent.
            q_update_target_estimator_every (int): The frequency of updating target network for
              inner DQN agent.
            q_discount_factor (float): The discount factor of inner DQN agent.
            q_epsilon_start (float): The starting epsilon of inner DQN agent.
            q_epsilon_end (float): the end epsilon of inner DQN agent.
            q_epsilon_decay_steps (int): The decay steps of inner DQN agent.
            q_batch_size (int): The batch size of inner DQN agent.
            q_train_step (int): Train the model every X steps.
            q_mlp_layers (list): The layer sizes of inner DQN agent.
            device (torch.device): Whether to use the cpu or gpu
        '''
        self.use_raw = False
        self._num_actions = num_actions
        self._state_shape = state_shape
        self._layer_sizes = hidden_layers_sizes + [num_actions]
        self._batch_size = batch_size
        self._train_every = train_every
        self._sl_learning_rate = sl_learning_rate
        self._anticipatory_param = anticipatory_param
        self._min_buffer_size_to_learn = min_buffer_size_to_learn

        self._reservoir_buffer = ReservoirBuffer(reservoir_buffer_capacity)
        self._prev_timestep = None
        self._prev_action = None
        self.evaluate_with = evaluate_with

        if device is None:
            self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = device

        # Total timesteps
        self.total_t = 0

        # Total training step
        self.train_t = 0

        # Build the action-value network
        self._rl_agent = DQNAgent(q_replay_memory_size, q_replay_memory_init_size, \
            q_update_target_estimator_every, q_discount_factor, q_epsilon_start, q_epsilon_end, \
            q_epsilon_decay_steps, q_batch_size, num_actions, state_shape, q_train_every, q_mlp_layers, \
            rl_learning_rate, device)

        # Build the average policy supervised model
        self._build_model()

        self.sample_episode_policy()
        
        # Checkpoint saving parameters
        self.save_path = save_path
        self.save_every = save_every

    def _build_model(self):
        ''' Build the average policy network
        '''

        # configure the average policy network
        policy_network = AveragePolicyNetwork(self._num_actions, self._state_shape, self._layer_sizes)
        policy_network = policy_network.to(self.device)
        self.policy_network = policy_network
        self.policy_network.eval()

        # xavier init
        for p in self.policy_network.parameters():
            if len(p.data.shape) > 1:
                nn.init.xavier_uniform_(p.data)

        # configure optimizer
        self.policy_network_optimizer = torch.optim.Adam(self.policy_network.parameters(), lr=self._sl_learning_rate)

    def feed(self, ts):
        ''' Feed data to inner RL agent

        Args:
            ts (list): A list of 5 elements that represent the transition.
        '''
        self._rl_agent.feed(ts)
        self.total_t += 1
        if self.total_t>0 and len(self._reservoir_buffer) >= self._min_buffer_size_to_learn and self.total_t%self._train_every == 0:
            sl_loss  = self.train_sl()
            print('\rINFO - Step {}, sl-loss: {}'.format(self.total_t, sl_loss), end='')

    def step(self, state):
        ''' Returns the action to be taken.

        Args:
            state (dict): The current state

        Returns:
            action (int): An action id
        '''
        obs = state['obs']
        legal_actions = list(state['legal_actions'].keys())
        if self._mode == 'best_response':
            action = self._rl_agent.step(state)
            one_hot = np.zeros(self._num_actions)
            one_hot[action] = 1
            self._add_transition(obs, one_hot)

        elif self._mode == 'average_policy':
            probs = self._act(obs)
            probs = remove_illegal(probs, legal_actions)
            action = np.random.choice(len(probs), p=probs)

        return action

    def eval_step(self, state):
        ''' Use the average policy for evaluation purpose

        Args:
            state (dict): The current state.

        Returns:
            action (int): An action id.
            info (dict): A dictionary containing information
        '''
        if self.evaluate_with == 'best_response':
            action, info = self._rl_agent.eval_step(state)
        elif self.evaluate_with == 'average_policy':
            obs = state['obs']
            legal_actions = list(state['legal_actions'].keys())
            probs = self._act(obs)
            probs = remove_illegal(probs, legal_actions)
            action = np.random.choice(len(probs), p=probs)
            info = {}
            info['probs'] = {state['raw_legal_actions'][i]: float(probs[list(state['legal_actions'].keys())[i]]) for i in range(len(state['legal_actions']))}
        else:
            raise ValueError("'evaluate_with' should be either 'average_policy' or 'best_response'.")
        return action, info

    def sample_episode_policy(self):
        ''' Sample average/best_response policy
        '''
        if np.random.rand() < self._anticipatory_param:
            self._mode = 'best_response'
        else:
            self._mode = 'average_policy'

    def _act(self, info_state):
        ''' Predict action probability givin the observation and legal actions
            Not connected to computation graph
        Args:
            info_state (numpy.array): An obervation.

        Returns:
            action_probs (numpy.array): The predicted action probability.
        '''
        info_state = np.expand_dims(info_state, axis=0)
        info_state = torch.from_numpy(info_state).float().to(self.device)

        with torch.no_grad():
            log_action_probs = self.policy_network(info_state).cpu().numpy()

        action_probs = np.exp(log_action_probs)[0]

        return action_probs

    def _add_transition(self, state, probs):
        ''' Adds the new transition to the reservoir buffer.

        Transitions are in the form (state, probs).

        Args:
            state (numpy.array): The state.
            probs (numpy.array): The probabilities of each action.
        '''
        transition = Transition(
                info_state=state,
                action_probs=probs)
        self._reservoir_buffer.add(transition)

    def train_sl(self):
        ''' Compute the loss on sampled transitions and perform a avg-network update.

        If there are not enough elements in the buffer, no loss is computed and
        `None` is returned instead.

        Returns:
            loss (float): The average loss obtained on this batch of transitions or `None`.
        '''
        if (len(self._reservoir_buffer) < self._batch_size or
                len(self._reservoir_buffer) < self._min_buffer_size_to_learn):
            return None

        transitions = self._reservoir_buffer.sample(self._batch_size)
        info_states = [t.info_state for t in transitions]
        action_probs = [t.action_probs for t in transitions]

        self.policy_network_optimizer.zero_grad()
        self.policy_network.train()

        # (batch, state_size)
        info_states = torch.from_numpy(np.array(info_states)).float().to(self.device)

        # (batch, num_actions)
        eval_action_probs = torch.from_numpy(np.array(action_probs)).float().to(self.device)

        # (batch, num_actions)
        log_forecast_action_probs = self.policy_network(info_states)

        ce_loss = - (eval_action_probs * log_forecast_action_probs).sum(dim=-1).mean()
        ce_loss.backward()

        self.policy_network_optimizer.step()
        ce_loss = ce_loss.item()
        self.policy_network.eval()

        self.train_t += 1

        if self.save_path and self.train_t % self.save_every == 0:
            # To preserve every checkpoint separately, 
            # add another argument to the function call parameterized by self.train_t
            self.save_checkpoint(self.save_path)
            print("\nINFO - Saved model checkpoint.")

        return ce_loss

    def set_device(self, device):
        self.device = device
        self._rl_agent.set_device(device)
        
    def checkpoint_attributes(self):
        '''
        Return the current checkpoint attributes (dict)
        Checkpoint attributes are used to save and restore the model in the middle of training
        Saves the model state dict, optimizer state dict, and all other instance variables
        '''
        
        return {
            'agent_type': 'NFSPAgent',
            'policy_network': self.policy_network.checkpoint_attributes(),
            'reservoir_buffer': self._reservoir_buffer.checkpoint_attributes(),
            'rl_agent': self._rl_agent.checkpoint_attributes(),
            'policy_network_optimizer': self.policy_network_optimizer.state_dict(),
            'device': self.device,
            'anticipatory_param': self._anticipatory_param,
            'batch_size': self._batch_size,
            'min_buffer_size_to_learn': self._min_buffer_size_to_learn,
            'num_actions': self._num_actions,
            'mode': self._mode,
            'evaluate_with': self.evaluate_with,
            'total_t': self.total_t,
            'train_t': self.train_t,
            'sl_learning_rate': self._sl_learning_rate,
            'train_every': self._train_every,
        }
    
    @classmethod
    def from_checkpoint(cls, checkpoint):
        '''
        Restore the model from a checkpoint
        
        Args:
            checkpoint (dict): the checkpoint attributes generated by checkpoint_attributes()
        '''
        print("\nINFO - Restoring model from checkpoint...")
        agent = cls(
            anticipatory_param=checkpoint['anticipatory_param'],
            batch_size=checkpoint['batch_size'],
            min_buffer_size_to_learn=checkpoint['min_buffer_size_to_learn'],
            num_actions=checkpoint['num_actions'],
            sl_learning_rate=checkpoint['sl_learning_rate'],
            train_every=checkpoint['train_every'],
            evaluate_with=checkpoint['evaluate_with'],
            device=checkpoint['device'],
            q_mlp_layers=checkpoint['rl_agent']['q_estimator']['mlp_layers'],
            state_shape=checkpoint['rl_agent']['q_estimator']['state_shape'],
            hidden_layers_sizes=[],
        )
        
        agent.policy_network = AveragePolicyNetwork.from_checkpoint(checkpoint['policy_network'])
        agent._reservoir_buffer = ReservoirBuffer.from_checkpoint(checkpoint['reservoir_buffer'])
        agent._mode = checkpoint['mode']
        agent.total_t = checkpoint['total_t']
        agent.train_t = checkpoint['train_t']
        agent.policy_network.to(agent.device)
        agent.policy_network.eval()
        agent.policy_network_optimizer = torch.optim.Adam(agent.policy_network.parameters(), lr=agent._sl_learning_rate)
        agent.policy_network_optimizer.load_state_dict(checkpoint['policy_network_optimizer'])
        agent._rl_agent.from_checkpoint(checkpoint['rl_agent'])
        agent._rl_agent.set_device(agent.device)
        return agent
        
    def save_checkpoint(self, path, filename='checkpoint_nfsp.pt'):
        ''' Save the model checkpoint (all attributes)

        Args:
            path (str): the path to save the model
        '''
        torch.save(self.checkpoint_attributes(), os.path.join(path, filename))
        

class AveragePolicyNetwork(nn.Module):
    '''
    Approximates the history of action probabilities
    given state (average policy). Forward pass returns
    log probabilities of actions.
    '''

    def __init__(self, num_actions=2, state_shape=None, mlp_layers=None):
        ''' Initialize the policy network.  It's just a bunch of ReLU
        layers with no activation on the final one, initialized with
        Xavier (sonnet.nets.MLP and tensorflow defaults)

        Args:
            num_actions (int): number of output actions
            state_shape (list): shape of state tensor for each sample
            mlp_laters (list): output size of each mlp layer including final
        '''
        super(AveragePolicyNetwork, self).__init__()

        self.num_actions = num_actions
        self.state_shape = state_shape
        self.mlp_layers = mlp_layers

        # set up mlp w/ relu activations
        layer_dims = [np.prod(self.state_shape)] + self.mlp_layers
        mlp = [nn.Flatten()]
        mlp.append(nn.BatchNorm1d(layer_dims[0]))
        for i in range(len(layer_dims)-1):
            mlp.append(nn.Linear(layer_dims[i], layer_dims[i+1]))
            if i != len(layer_dims) - 2: # all but final have relu
                mlp.append(nn.ReLU())
        self.mlp = nn.Sequential(*mlp)

    def forward(self, s):
        ''' Log action probabilities of each action from state

        Args:
            s (Tensor): (batch, state_shape) state tensor

        Returns:
            log_action_probs (Tensor): (batch, num_actions)
        '''
        logits = self.mlp(s)
        log_action_probs = F.log_softmax(logits, dim=-1)
        return log_action_probs
    
    def checkpoint_attributes(self):
        '''
        Return the current checkpoint attributes (dict)
        Checkpoint attributes are used to save and restore the model in the middle of training
        '''
        
        return {
            'num_actions': self.num_actions,
            'state_shape': self.state_shape,
            'mlp_layers': self.mlp_layers,
            'mlp': self.mlp.state_dict(),
        }
        
    @classmethod
    def from_checkpoint(cls, checkpoint):
        '''
        Restore the model from a checkpoint
        
        Args:
            checkpoint (dict): the checkpoint attributes generated by checkpoint_attributes()
        '''
        
        agent = cls(
            num_actions=checkpoint['num_actions'],
            state_shape=checkpoint['state_shape'],
            mlp_layers=checkpoint['mlp_layers'],
        )
        
        agent.mlp.load_state_dict(checkpoint['mlp'])
        return agent

class ReservoirBuffer(object):
    ''' Allows uniform sampling over a stream of data.

    This class supports the storage of arbitrary elements, such as observation
    tensors, integer actions, etc.

    See https://en.wikipedia.org/wiki/Reservoir_sampling for more details.
    '''

    def __init__(self, reservoir_buffer_capacity):
        ''' Initialize the buffer.
        '''
        self._reservoir_buffer_capacity = reservoir_buffer_capacity
        self._data = []
        self._add_calls = 0

    def add(self, element):
        ''' Potentially adds `element` to the reservoir buffer.

        Args:
            element (object): data to be added to the reservoir buffer.
        '''
        if len(self._data) < self._reservoir_buffer_capacity:
            self._data.append(element)
        else:
            idx = np.random.randint(0, self._add_calls + 1)
            if idx < self._reservoir_buffer_capacity:
                self._data[idx] = element
        self._add_calls += 1

    def sample(self, num_samples):
        ''' Returns `num_samples` uniformly sampled from the buffer.

        Args:
            num_samples (int): The number of samples to draw.

        Returns:
            An iterable over `num_samples` random elements of the buffer.

        Raises:
            ValueError: If there are less than `num_samples` elements in the buffer
        '''
        if len(self._data) < num_samples:
            raise ValueError("{} elements could not be sampled from size {}".format(
                    num_samples, len(self._data)))
        return random.sample(self._data, num_samples)

    def clear(self):
        ''' Clear the buffer
        '''
        self._data = []
        self._add_calls = 0
        
    def checkpoint_attributes(self):
        return {
            'data': self._data,
            'add_calls': self._add_calls,
            'reservoir_buffer_capacity': self._reservoir_buffer_capacity,
        }
        
    @classmethod
    def from_checkpoint(cls, checkpoint):
        reservoir_buffer = cls(checkpoint['reservoir_buffer_capacity'])
        reservoir_buffer._data = checkpoint['data']
        reservoir_buffer._add_calls = checkpoint['add_calls']
        return reservoir_buffer

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)




'''
    Project: Gui Gin Rummy
    File name: canvas_item.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from rlcard.agents.gin_rummy_human_agent.gui_gin_rummy.game_canvas import GameCanvas
    from rlcard.agents.gin_rummy_human_agent.gui_cards.card_image import CardImage


class CanvasItem(object):

    def __init__(self, item_id: int, game_canvas: 'GameCanvas'):
        self.item_id = item_id
        self.game_canvas = game_canvas

    def __eq__(self, other):
        if isinstance(other, int):  # FIXME: temporary kludge to convert all item_id to CanvasItem
            return other == self.item_id
        return isinstance(other, CanvasItem) and self.item_id == other.item_id

    def __hash__(self):
        return hash(self.item_id)

    def get_tags(self):
        return self.game_canvas.gettags(self.item_id)


class CardItem(CanvasItem):

    def __init__(self, item_id: int, card_id: int, card_image: 'CardImage', game_canvas: 'GameCanvas'):
        super().__init__(item_id=item_id, game_canvas=game_canvas)
        self.card_id = card_id
        self.card_image = card_image

    def is_face_up(self) -> bool:
        return self.card_image.face_up

    def set_card_id_face_up(self, face_up: bool):
        if self.card_image.face_up != face_up:
            target_image = self.card_image if face_up else self.game_canvas.card_back_image
            self.game_canvas.itemconfig(self.item_id, image=target_image)
            self.card_image.face_up = face_up

    def flip_over(self):
        self.card_image.face_up = not self.card_image.face_up
        target_image = self.card_image if self.card_image.face_up else self.game_canvas.card_back_image
        self.game_canvas.itemconfig(self.item_id, image=target_image)



from rlcard.games.uno.card import UnoCard

class HumanAgent(object):
    ''' A human agent for Leduc Holdem. It can be used to play against trained models
    '''

    def __init__(self, num_actions):
        ''' Initilize the human agent

        Args:
            num_actions (int): the size of the ouput action space
        '''
        self.use_raw = True
        self.num_actions = num_actions

    @staticmethod
    def step(state):
        ''' Human agent will display the state and make decisions through interfaces

        Args:
            state (dict): A dictionary that represents the current state

        Returns:
            action (int): The action decided by human
        '''
        print(state['raw_obs'])
        _print_state(state['raw_obs'], state['action_record'])
        action = int(input('>> You choose action (integer): '))
        while action < 0 or action >= len(state['legal_actions']):
            print('Action illegal...')
            action = int(input('>> Re-choose action (integer): '))
        return state['raw_legal_actions'][action]

    def eval_step(self, state):
        ''' Predict the action given the curent state for evaluation. The same to step here.

        Args:
            state (numpy.array): an numpy array that represents the current state

        Returns:
            action (int): the action predicted (randomly chosen) by the random agent
        '''
        return self.step(state), {}

def _print_state(state, action_record):
    ''' Print out the state of a given player

    Args:
        player (int): Player id
    '''
    _action_list = []
    for i in range(1, len(action_record)+1):
        if action_record[-i][0] == state['current_player']:
            break
        _action_list.insert(0, action_record[-i])
    for pair in _action_list:
        print('>> Player', pair[0], 'chooses ', end='')
        _print_action(pair[1])
        print('')

    print('\n=============== Your Hand ===============')
    UnoCard.print_cards(state['hand'])
    print('')
    print('=============== Last Card ===============')
    UnoCard.print_cards(state['target'], wild_color=True)
    print('')
    print('========== Players Card Number ===========')
    for i in range(state['num_players']):
        if i != state['current_player']:
            print('Player {} has {} cards.'.format(i, state['num_cards'][i]))
    print('======== Actions You Can Choose =========')
    for i, action in enumerate(state['legal_actions']):
        print(str(i)+': ', end='')
        UnoCard.print_cards(action, wild_color=True)
        if i < len(state['legal_actions']) - 1:
            print(', ', end='')
    print('\n')

def _print_action(action):
    ''' Print out an action in a nice form

    Args:
        action (str): A string a action
    '''
    UnoCard.print_cards(action, wild_color=True)



from copy import deepcopy
import numpy as np

from rlcard.games.blackjack import Dealer
from rlcard.games.blackjack import Player
from rlcard.games.blackjack import Judger

class BlackjackGame:

    def __init__(self, allow_step_back=False):
        ''' Initialize the class Blackjack Game
        '''
        self.allow_step_back = allow_step_back
        self.np_random = np.random.RandomState()

    def configure(self, game_config):
        ''' Specifiy some game specific parameters, such as number of players
        '''
        self.num_players = game_config['game_num_players']
        self.num_decks = game_config['game_num_decks']

    def init_game(self):
        ''' Initialilze the game

        Returns:
            state (dict): the first state of the game
            player_id (int): current player's id
        '''
        self.dealer = Dealer(self.np_random, self.num_decks)

        self.players = []
        for i in range(self.num_players):
            self.players.append(Player(i, self.np_random))

        self.judger = Judger(self.np_random)

        for i in range(2):
            for j in range(self.num_players):
                self.dealer.deal_card(self.players[j])
            self.dealer.deal_card(self.dealer)

        for i in range(self.num_players):
            self.players[i].status, self.players[i].score = self.judger.judge_round(self.players[i])

        self.dealer.status, self.dealer.score = self.judger.judge_round(self.dealer)

        self.winner = {'dealer': 0}
        for i in range(self.num_players):
            self.winner['player' + str(i)] = 0

        self.history = []
        self.game_pointer = 0

        return self.get_state(self.game_pointer), self.game_pointer

    def step(self, action):
        ''' Get the next state

        Args:
            action (str): a specific action of blackjack. (Hit or Stand)

        Returns:/
            dict: next player's state
            int: next plater's id
        '''
        if self.allow_step_back:
            p = deepcopy(self.players[self.game_pointer])
            d = deepcopy(self.dealer)
            w = deepcopy(self.winner)
            self.history.append((d, p, w))

        next_state = {}
        # Play hit
        if action != "stand":
            self.dealer.deal_card(self.players[self.game_pointer])
            self.players[self.game_pointer].status, self.players[self.game_pointer].score = self.judger.judge_round(
                self.players[self.game_pointer])
            if self.players[self.game_pointer].status == 'bust':
                # game over, set up the winner, print out dealer's hand # If bust, pass the game pointer
                if self.game_pointer >= self.num_players - 1:
                    while self.judger.judge_score(self.dealer.hand) < 17:
                        self.dealer.deal_card(self.dealer)
                    self.dealer.status, self.dealer.score = self.judger.judge_round(self.dealer)
                    for i in range(self.num_players):
                        self.judger.judge_game(self, i) 
                    self.game_pointer = 0
                else:
                    self.game_pointer += 1

                
        elif action == "stand": # If stand, first try to pass the pointer, if it's the last player, dealer deal for himself, then judge game for everyone using a loop
            self.players[self.game_pointer].status, self.players[self.game_pointer].score = self.judger.judge_round(
                self.players[self.game_pointer])
            if self.game_pointer >= self.num_players - 1:
                while self.judger.judge_score(self.dealer.hand) < 17:
                    self.dealer.deal_card(self.dealer)
                self.dealer.status, self.dealer.score = self.judger.judge_round(self.dealer)
                for i in range(self.num_players):
                    self.judger.judge_game(self, i) 
                self.game_pointer = 0
            else:
                self.game_pointer += 1


            
            

        hand = [card.get_index() for card in self.players[self.game_pointer].hand]

        if self.is_over():
            dealer_hand = [card.get_index() for card in self.dealer.hand]
        else:
            dealer_hand = [card.get_index() for card in self.dealer.hand[1:]]

        for i in range(self.num_players):
            next_state['player' + str(i) + ' hand'] = [card.get_index() for card in self.players[i].hand]
        next_state['dealer hand'] = dealer_hand
        next_state['actions'] = ('hit', 'stand')
        next_state['state'] = (hand, dealer_hand)

        

        return next_state, self.game_pointer

    def step_back(self):
        ''' Return to the previous state of the game

        Returns:
            Status (bool): check if the step back is success or not
        '''
        #while len(self.history) > 0:
        if len(self.history) > 0:
            self.dealer, self.players[self.game_pointer], self.winner = self.history.pop()
            return True
        return False

    def get_num_players(self):
        ''' Return the number of players in blackjack

        Returns:
            number_of_player (int): blackjack only have 1 player
        '''
        return self.num_players

    @staticmethod
    def get_num_actions():
        ''' Return the number of applicable actions

        Returns:
            number_of_actions (int): there are only two actions (hit and stand)
        '''
        return 2

    def get_player_id(self):
        ''' Return the current player's id

        Returns:
            player_id (int): current player's id
        '''
        return self.game_pointer

    def get_state(self, player_id):
        ''' Return player's state

        Args:
            player_id (int): player id

        Returns:
            state (dict): corresponding player's state
        '''
        '''
                before change state only have two keys (action, state)
                but now have more than 4 keys (action, state, player0 hand, player1 hand, ... , dealer hand)
                Although key 'state' have duplicated information with key 'player hand' and 'dealer hand', I couldn't remove it because of other codes
                To remove it, we need to change dqn agent too in my opinion
                '''
        state = {}
        state['actions'] = ('hit', 'stand')
        hand = [card.get_index() for card in self.players[player_id].hand]
        if self.is_over():
            dealer_hand = [card.get_index() for card in self.dealer.hand]
        else:
            dealer_hand = [card.get_index() for card in self.dealer.hand[1:]]

        for i in range(self.num_players):
            state['player' + str(i) + ' hand'] = [card.get_index() for card in self.players[i].hand]
        state['dealer hand'] = dealer_hand
        state['state'] = (hand, dealer_hand)

        return state

    def is_over(self):
        ''' Check if the game is over

        Returns:
            status (bool): True/False
        '''
        '''
                I should change here because judger and self.winner is changed too
                '''
        for i in range(self.num_players):
            if self.winner['player' + str(i)] == 0:
                return False

        return True



import unittest

from rlcard.agents.human_agents.uno_human_agent import _print_state, _print_action

class TestLeducHuman(unittest.TestCase):

    def test_print_state(self):
        raw_state = {'target': 'r-reverse', 'current_player': 0, 'legal_actions': ['r-2', 'r-draw_2'], 'hand': ['y-skip', 'y-draw_2', 'r-2', 'b-3', 'b-6', 'g-wild_draw_4', 'r-draw_2'], 'played_cards': ['g-reverse', 'r-reverse'], 'num_players': 2, 'others_hand': ['y-4', 'g-6', 'b-reverse', 'b-5', 'b-reverse', 'r-9'], 'num_cards': [7, 6]}
        action_record = []
        _print_state(raw_state, action_record)

    def test_print_action(self):
        _print_action('r-8')



# Copyright 2021 RLCard Team of Texas A&M University
# Copyright 2021 DouZero Team of Kwai
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import torch
from torch import nn

class DMCNet(nn.Module):
    def __init__(
        self,
        state_shape,
        action_shape,
        mlp_layers=[512,512,512,512,512]
    ):
        super().__init__()
        input_dim = np.prod(state_shape) + np.prod(action_shape)
        layer_dims = [input_dim] + mlp_layers
        fc = []
        for i in range(len(layer_dims)-1):
            fc.append(nn.Linear(layer_dims[i], layer_dims[i+1]))
            fc.append(nn.ReLU())
        fc.append(nn.Linear(layer_dims[-1], 1))
        self.fc_layers = nn.Sequential(*fc)

    def forward(self, obs, actions):
        obs = torch.flatten(obs, 1)
        actions = torch.flatten(actions, 1)
        x = torch.cat((obs, actions), dim=1)
        values = self.fc_layers(x).flatten()
        return values

class DMCAgent:
    def __init__(
        self,
        state_shape,
        action_shape,
        mlp_layers=[512,512,512,512,512],
        exp_epsilon=0.01,
        device="0",
    ):
        self.use_raw = False
        self.device = 'cuda:'+device if device != "cpu" else "cpu"
        self.net = DMCNet(state_shape, action_shape, mlp_layers).to(self.device)
        self.exp_epsilon = exp_epsilon
        self.action_shape = action_shape

    def step(self, state):
        action_keys, values = self.predict(state)

        if self.exp_epsilon > 0 and np.random.rand() < self.exp_epsilon:
            action = np.random.choice(action_keys)
        else:
            action_idx = np.argmax(values)
            action = action_keys[action_idx]

        return action

    def eval_step(self, state):
        action_keys, values = self.predict(state)

        action_idx = np.argmax(values)
        action = action_keys[action_idx]

        info = {}
        info['values'] = {state['raw_legal_actions'][i]: float(values[i]) for i in range(len(action_keys))}

        return action, info

    def share_memory(self):
        self.net.share_memory()

    def eval(self):
        self.net.eval()

    def parameters(self):
        return self.net.parameters()

    def predict(self, state):
        # Prepare obs and actions
        obs = state['obs'].astype(np.float32)
        legal_actions = state['legal_actions']
        action_keys = np.array(list(legal_actions.keys()))
        action_values = list(legal_actions.values())
        # One-hot encoding if there is no action features
        for i in range(len(action_values)):
            if action_values[i] is None:
                action_values[i] = np.zeros(self.action_shape[0])
                action_values[i][action_keys[i]] = 1
        action_values = np.array(action_values, dtype=np.float32)

        obs = np.repeat(obs[np.newaxis, :], len(action_keys), axis=0)

        # Predict Q values
        values = self.net.forward(torch.from_numpy(obs).to(self.device),
                                  torch.from_numpy(action_values).to(self.device))

        return action_keys, values.cpu().detach().numpy()

    def forward(self, obs, actions):
        return self.net.forward(obs, actions)

    def load_state_dict(self, state_dict):
        return self.net.load_state_dict(state_dict)

    def state_dict(self):
        return self.net.state_dict()

    def set_device(self, device):
        self.device = device

class DMCModel:
    def __init__(
        self,
        state_shape,
        action_shape,
        mlp_layers=[512,512,512,512,512],
        exp_epsilon=0.01,
        device=0
    ):
        self.agents = []
        for player_id in range(len(state_shape)):
            agent = DMCAgent(
                state_shape[player_id],
                action_shape[player_id],
                mlp_layers,
                exp_epsilon,
                device,
            )
            self.agents.append(agent)

    def share_memory(self):
        for agent in self.agents:
            agent.share_memory()

    def eval(self):
        for agent in self.agents:
            agent.eval()

    def parameters(self, index):
        return self.agents[index].parameters()

    def get_agent(self, index):
        return self.agents[index]

    def get_agents(self):
        return self.agents



'''
    File name: bridge/utils/utils.py
    Author: William Hale
    Date created: 11/26/2021
'''

from typing import List

import numpy as np

from .bridge_card import BridgeCard


def encode_cards(cards: List[BridgeCard]) -> np.ndarray:  # Note: not used ??
    plane = np.zeros(52, dtype=int)
    for card in cards:
        plane[card.card_id] = 1
    return plane



'''
    Project: Gui Gin Rummy
    File name: menu_bar.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_frame import GameFrame

import tkinter as tk
from tkinter import messagebox

from .preferences_window import PreferencesWindow


class MenuBar(tk.Menu):

    def __init__(self, root: tk.Tk, game_frame: 'GameFrame'):
        super().__init__(root)
        self.game_frame = game_frame

        # create file menu
        file_menu = tk.Menu(self, tearoff=False)
        file_menu.add_command(label="New Game", command=self.on_new_game_menu_clicked)
        self.add_cascade(label="File", menu=file_menu)

        # create edit menu
        edit_menu = tk.Menu(self, tearoff=False)
        edit_menu.add_command(label="Preferences", command=self.on_preference_menu_clicked)
        self.add_cascade(label="Edit", menu=edit_menu)

        # create about menu
        help_menu = tk.Menu(self, tearoff=False)
        help_menu.add_command(label="About", command=self.on_about_menu_clicked)
        self.add_cascade(label="Help", menu=help_menu)

        # configure menuBar
        root.configure(menu=self)

    def on_new_game_menu_clicked(self):
        self.game_frame.start_new_game()

    def on_preference_menu_clicked(self):
        PreferencesWindow(self.game_frame)

    @staticmethod
    def on_about_menu_clicked():
        messagebox.showinfo(title="Info", message="Gin Rummy\nVersion 1.0")



''' Wrrapers of pretrained models.
'''
import os

import rlcard
from rlcard.agents import CFRAgent
from rlcard.models.model import Model

# Root path of pretrianed models
ROOT_PATH = os.path.join(rlcard.__path__[0], 'models/pretrained')

class LeducHoldemCFRModel(Model):
    ''' A pretrained model on Leduc Holdem with CFR (chance sampling)
    '''
    def __init__(self):
        ''' Load pretrained model
        '''
        env = rlcard.make('leduc-holdem')
        self.agent = CFRAgent(env, model_path=os.path.join(ROOT_PATH, 'leduc_holdem_cfr'))
        self.agent.load()
    @property
    def agents(self):
        ''' Get a list of agents for each position in a the game

        Returns:
            agents (list): A list of agents

        Note: Each agent should be just like RL agent with step and eval_step
              functioning well.
        '''
        return [self.agent, self.agent]




import numpy as np


class RandomAgent(object):
    ''' A random agent. Random agents is for running toy examples on the card games
    '''

    def __init__(self, num_actions):
        ''' Initilize the random agent

        Args:
            num_actions (int): The size of the ouput action space
        '''
        self.use_raw = False
        self.num_actions = num_actions

    @staticmethod
    def step(state):
        ''' Predict the action given the curent state in gerenerating training data.

        Args:
            state (dict): An dictionary that represents the current state

        Returns:
            action (int): The action predicted (randomly chosen) by the random agent
        '''
        return np.random.choice(list(state['legal_actions'].keys()))

    def eval_step(self, state):
        ''' Predict the action given the current state for evaluation.
            Since the random agents are not trained. This function is equivalent to step function

        Args:
            state (dict): An dictionary that represents the current state

        Returns:
            action (int): The action predicted (randomly chosen) by the random agent
            probs (list): The list of action probabilities
        '''
        probs = [0 for _ in range(self.num_actions)]
        for i in state['legal_actions']:
            probs[i] = 1/len(state['legal_actions'])

        info = {}
        info['probs'] = {state['raw_legal_actions'][i]: probs[list(state['legal_actions'].keys())[i]] for i in range(len(state['legal_actions']))}

        return self.step(state), info
